import asyncio
import io
import os
import secrets
import sys
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum, auto
from typing import (
    TYPE_CHECKING,
    ContextManager,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    NoReturn,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

import aiohttp
from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.parser_helpers import read_exact
from lonelypsp.util.async_queue_like import AsyncQueueLike
from lonelypsp.util.drainable_asyncio_queue import DrainableAsyncioQueue, QueueDrained

from lonelypsc.client import (
    PubSubClient,
    PubSubClientBulkSubscriptionConnector,
    PubSubClientConnectionStatus,
    PubSubClientConnector,
    PubSubClientMessageWithCleanup,
    PubSubClientReceiver,
    PubSubClientTracingNotifyOnHashed,
    PubSubClientTracingNotifyStart,
    PubSubDirectConnectionStatusReceiver,
    PubSubDirectOnMessageWithCleanupReceiver,
    PubSubNotifyResult,
    PubSubRequestAmbiguousError,
    PubSubRequestConnectionAbandonedError,
)
from lonelypsc.config.config import BroadcastersShuffler
from lonelypsc.config.ws_config import WebsocketPubSubConfig
from lonelypsc.http_client import HttpPubSubNotifyResult
from lonelypsc.types.sync_io import (
    SyncStandardIO,
)
from lonelypsc.util.errors import combine_multiple_exceptions
from lonelypsc.util.io_helpers import PositionedSyncStandardIO
from lonelypsc.ws.handlers.handler import handle_any
from lonelypsc.ws.state import (
    ClosingRetryInformationType,
    InternalLargeMessage,
    InternalMessage,
    InternalMessageState,
    InternalMessageStateAndCallback,
    InternalMessageStateType,
    InternalMessageStateUnsent,
    InternalMessageType,
    InternalSmallMessage,
    ManagementTask,
    ManagementTaskSubscribeExact,
    ManagementTaskSubscribeGlob,
    ManagementTaskType,
    ManagementTaskUnsubscribeExact,
    ManagementTaskUnsubscribeGlob,
    ReceivedLargeMessage,
    ReceivedMessage,
    ReceivedMessageType,
    ReceivedSmallMessage,
    RetryInformation,
    State,
    StateClosed,
    StateConnecting,
    StateType,
    TasksOnceOpen,
)
from lonelypsc.ws.websocket_connect_task import make_websocket_connect_task

if sys.version_info >= (3, 11):
    from typing import Never
else:
    from typing import NoReturn as Never


class NoopStatelessTracingNotifyStart:
    def on_start_without_hash(
        self, /, *, topic: bytes, length: int, filelike: bool
    ) -> PubSubClientTracingNotifyOnHashed[Literal[None]]:
        return self

    def on_hashed(self) -> None: ...

    def on_start_with_hash(
        self, /, *, topic: bytes, length: int, filelike: bool
    ) -> None: ...


@contextmanager
def _noop_notify_tracer() -> Iterator[NoopStatelessTracingNotifyStart]:
    yield NoopStatelessTracingNotifyStart()


@fast_dataclass
class WsPubSubNotifyResult:
    notified: int


class CRStateType(Enum):
    NOT_SETUP = auto()
    SETUP = auto()
    ERRORED = auto()
    TORN_DOWN = auto()


@dataclass
class CRStateNotSetup:
    """The state variables before a setup call"""

    type: Literal[CRStateType.NOT_SETUP]
    """discriminator value"""

    config: WebsocketPubSubConfig
    """how the subscriber is configured"""

    receiver_counter: int
    """the id to assign to the next message/status receiver"""

    message_receivers: Dict[int, PubSubDirectOnMessageWithCleanupReceiver]
    """who to inform when the subscriber receives a message on a topic

    since the subscriber is not yet setup, these are not called yet
    """

    status_receivers: Dict[int, PubSubDirectConnectionStatusReceiver]
    """who to inform if the connection status changes. these are expected to be
    under the assumption the connection status is currently LOST.
    
    since the subscriber is not yet setup, these do not need to be called until
    at least setup
    """


@dataclass
class CRStateSetup:
    """The state variables after a setup call and before any teardown call"""

    type: Literal[CRStateType.SETUP]
    """discriminator value"""

    config: WebsocketPubSubConfig
    """how the subscriber is configured"""

    receiver_counter: int
    """the id to assign to the next message/status receiver"""

    message_receivers: Dict[int, PubSubDirectOnMessageWithCleanupReceiver]
    """who to inform when the subscriber receives a message on a topic
    
    called within the state task

    if mutated during a call: every entry that is in this for the entire
    iteration will be called, and no entry will be called more than once
    """

    new_status_receivers: DrainableAsyncioQueue[
        Tuple[int, PubSubDirectConnectionStatusReceiver]
    ]
    """status receivers which have not yet been put into status_receivers because they
    have not yet been informed about the current connection status, which needs to be
    done within the state_task
    """

    removed_status_receivers: DrainableAsyncioQueue[int]
    """status receivers which may be in new_status_receivers or status_receivers but
    need to be removed from there, which must be done within the state_task for
    simplicity and symmetry
    """

    status_receivers: Dict[int, PubSubDirectConnectionStatusReceiver]
    """who to inform if the connection status changes.
    
    called within the state task

    if mutated during a call: every entry that is in this for the entire
    iteration will be called, and no entry will be called more than once
    """

    connection_lost_flag: bool
    """A flag that can be set to true to indicate the state task should
    consider the connection LOST if it were to be considered OK in the next
    iteration
    """

    ws_state: State
    """the state of the connection to the broadcasters, mutated by the state task.
    unless otherwise noted it should be assumed all inner variables are not asyncio
    safe
    """

    state_task: asyncio.Task[None]
    """the task that is mutating ws_state"""


@fast_dataclass
class CRStateErrored:
    """the state variables if the state task errors/finishes unexpectedly and before
    any teardown call
    """

    type: Literal[CRStateType.ERRORED]
    """discriminator value"""

    exception: BaseException
    """the exception to raise when trying to use the connection"""


@fast_dataclass
class CRStateTornDown:
    """the state variables after a teardown call"""

    type: Literal[CRStateType.TORN_DOWN]
    """discriminator value"""


CRState = Union[
    CRStateNotSetup,
    CRStateSetup,
    CRStateErrored,
    CRStateTornDown,
]


class WSPubSubConnectorReceiver:
    """The connector/receiver for the websocket client; unlike with the
    http client, both outgoing and incoming messages are handled within
    the same connection, so splitting the connector and receiver is less
    useful

    In order to maintain the same signature as the standard connector/receiver,
    this will treat setting up the connector or receiver as setting up this
    object, then tearing down either as tearing down this object, and ignore
    repeated setups/teardowns02
    """

    def __init__(self, config: WebsocketPubSubConfig) -> None:
        self.state: CRState = CRStateNotSetup(
            type=CRStateType.NOT_SETUP,
            config=config,
            receiver_counter=1,
            message_receivers=dict(),
            status_receivers=dict(),
        )
        """our state; we add a layer of indirection to make the relationships between
        the state variables more clear
        """

    async def _setup(self) -> None:
        """The implementation for setup connector / setup receiver"""
        if self.state.type != CRStateType.NOT_SETUP:
            if self.state.type == CRStateType.ERRORED:
                raise self.state.exception
            if self.state.type == CRStateType.TORN_DOWN:
                raise Exception("cannot setup after teardown")
            return

        shuffler = BroadcastersShuffler(self.state.config.broadcasters)
        retry = RetryInformation(
            shuffler=shuffler,
            iteration=0,
            iterator=iter(shuffler),
        )
        broadcaster = next(retry.iterator)

        client_session = aiohttp.ClientSession()
        self.state = CRStateSetup(
            type=CRStateType.SETUP,
            config=self.state.config,
            receiver_counter=self.state.receiver_counter,
            message_receivers=self.state.message_receivers,
            status_receivers=self.state.status_receivers,
            new_status_receivers=DrainableAsyncioQueue(),
            removed_status_receivers=DrainableAsyncioQueue(),
            connection_lost_flag=False,
            ws_state=StateConnecting(
                type=StateType.CONNECTING,
                config=self.state.config,
                client_session=client_session,
                websocket_task=make_websocket_connect_task(
                    config=self.state.config,
                    broadcaster=broadcaster,
                    client_session=client_session,
                ),
                cancel_requested=asyncio.Event(),
                broadcaster=broadcaster,
                retry=retry,
                tasks=TasksOnceOpen(
                    exact_subscriptions=set(),
                    glob_subscriptions=set(),
                    unsorted=DrainableAsyncioQueue(
                        max_size=self.state.config.max_expected_acks
                    ),
                    unsent_notifications=DrainableAsyncioQueue(
                        max_size=self.state.config.max_unsent_notifications
                    ),
                    resending_notifications=[],
                ),
                backgrounded=set(),
            ),
            state_task=cast(asyncio.Task[None], None),  # order of initialization
        )
        self.state.state_task = asyncio.create_task(self._state_task_target())

    async def _teardown(self) -> None:
        """The implementation for teardown connector / teardown receiver"""
        if (
            self.state.type == CRStateType.NOT_SETUP
            or self.state.type == CRStateType.ERRORED
            or self.state.type == CRStateType.TORN_DOWN
        ):
            self.state = CRStateTornDown(type=CRStateType.TORN_DOWN)
            return

        if self.state.ws_state.type != StateType.CLOSED:
            self.state.ws_state.cancel_requested.set()
        try:
            await self.state.state_task
        finally:
            self.state = CRStateTornDown(type=CRStateType.TORN_DOWN)

    async def _state_task_target(self) -> None:
        """the target for state_task when in StateType.SETUP; manages actually looping the
        websocket state and notifying message/status receivers based on changes to that state

        this will complete when the state moves away from SETUP or the websocket state
        moves to closed. Thus, to cancel this task, do not use standard cancellation, instead,
        set cancel_requested on the websocket state which will tell the underlying handler to
        gracefully shutdown
        """
        state = self.state
        if state.type != CRStateType.SETUP:
            return
        receivers_expect = PubSubClientConnectionStatus.LOST
        receiver_errors: List[BaseException] = []
        handler_error: Optional[BaseException] = None
        received_queue: DrainableAsyncioQueue[ReceivedMessage] = DrainableAsyncioQueue(
            max_size=state.config.max_received
        )
        received_missed: asyncio.Event = asyncio.Event()
        received_task = asyncio.create_task(
            self._handle_received_task_target(received_queue, received_missed)
        )

        while True:
            state = self.state
            if state.type != CRStateType.SETUP:
                break

            new_receivers = dict(state.new_status_receivers.drain())
            state.new_status_receivers = DrainableAsyncioQueue()

            state.status_receivers.update(new_receivers)

            removed_receivers = state.removed_status_receivers.drain()
            state.removed_status_receivers = DrainableAsyncioQueue()

            for receiver_id in removed_receivers:
                state.status_receivers.pop(receiver_id, None)
                new_receivers.pop(receiver_id, None)

            if receivers_expect != PubSubClientConnectionStatus.LOST:
                await self._inform_status_receivers(
                    receivers_expect, new_receivers.values(), receiver_errors
                )

                if receiver_errors and state.ws_state.type != StateType.CLOSED:
                    state.ws_state.cancel_requested.set()

            old_received = (
                None
                if state.ws_state.type != StateType.OPEN
                else state.ws_state.received
            )
            try:
                state.ws_state = await handle_any(state.ws_state)
            except BaseException as e:
                handler_error = e
                state.ws_state = StateClosed(type=StateType.CLOSED)

            connection_statuses: List[PubSubClientConnectionStatus] = [
                PubSubClientConnectionStatus.LOST
            ]

            if received_task.done():
                connection_statuses = [PubSubClientConnectionStatus.ABANDONED]
                exc = received_task.exception()
                if exc is not None:
                    receiver_errors.append(exc)
                else:
                    receiver_errors.append(
                        Exception("received_task finished unexpectedly")
                    )
            elif (
                state.ws_state.type != StateType.CLOSED
                and state.ws_state.cancel_requested.is_set()
            ):
                connection_statuses = [PubSubClientConnectionStatus.ABANDONED]
            elif (
                state.ws_state.type == StateType.OPEN
                # if there are no unsent management tasks and all the acks can be explained by sent
                # notifications, then all subscriptions/unsubscriptions have been ack'd
                and state.ws_state.management_tasks.qsize() == 0
                and len(state.ws_state.expected_acks)
                <= len(state.ws_state.sent_notifications)
            ):
                if state.connection_lost_flag or received_missed.is_set():
                    connection_statuses = [
                        PubSubClientConnectionStatus.LOST,
                        PubSubClientConnectionStatus.OK,
                    ]
                else:
                    connection_statuses = [PubSubClientConnectionStatus.OK]
            elif state.ws_state.type == StateType.CLOSED:
                connection_statuses = [PubSubClientConnectionStatus.ABANDONED]

            state.connection_lost_flag = False
            received_missed.clear()

            for connection_status in connection_statuses:
                if connection_status == receivers_expect:
                    continue
                await self._inform_status_receivers(
                    connection_status,
                    state.status_receivers.values(),
                    receiver_errors,
                )
                receivers_expect = connection_status

                if not receiver_errors:
                    continue

                if receivers_expect != PubSubClientConnectionStatus.ABANDONED:
                    await self._inform_status_receivers(
                        PubSubClientConnectionStatus.ABANDONED,
                        state.status_receivers.values(),
                        receiver_errors,
                    )

                if state.ws_state.type != StateType.CLOSED:
                    state.ws_state.cancel_requested.set()

                receivers_expect = PubSubClientConnectionStatus.ABANDONED
                break

            if state.ws_state.type != StateType.OPEN:
                if old_received is not None:
                    for message in old_received.drain():
                        received_queue.put_nowait(message)
            else:
                while True:
                    try:
                        message = state.ws_state.received.get_nowait()
                    except asyncio.QueueEmpty:
                        break

                    received_queue.put_nowait(message)

            if state.ws_state.type == StateType.CLOSED:
                should_process = not received_task.done()
                unprocessed = received_queue.drain()
                try:
                    await received_task
                except BaseException as e:
                    receiver_errors.append(e)

                if should_process:
                    old_error_count = len(receiver_errors)
                    for message in unprocessed:
                        if message.type == ReceivedMessageType.MISSED:
                            continue

                        await self._handle_received_message(
                            message, state.message_receivers.values(), receiver_errors
                        )
                        if len(receiver_errors) > old_error_count:
                            break
                break

        if receiver_errors:
            raise combine_multiple_exceptions(
                "handlers raised error", receiver_errors, context=handler_error
            )

        if handler_error:
            raise handler_error

    async def _inform_status_receivers(
        self,
        status: PubSubClientConnectionStatus,
        receivers: Iterable[PubSubDirectConnectionStatusReceiver],
        errors: List[BaseException],
    ) -> None:
        """informs each of the status receivers about the new status; if any of them
        raise errors (including asyncio.CancelledError), puts those errors in the given
        list to be merged and raised at the appropriate point. also, critically, this
        ensures that a receiver that is later in the list is still informed if an earlier
        one raises an exception
        """
        for receiver in receivers:
            try:
                if status == PubSubClientConnectionStatus.OK:
                    await receiver.on_connection_established()
                elif status == PubSubClientConnectionStatus.LOST:
                    await receiver.on_connection_lost()
                else:
                    # verify type
                    _: Literal[PubSubClientConnectionStatus.ABANDONED] = status
                    await receiver.on_connection_abandoned()
            except BaseException as e:
                errors.append(e)

    async def _handle_received_task_target(
        self, queue: AsyncQueueLike[ReceivedMessage], received_missed: asyncio.Event
    ) -> None:
        """target for received_task within _state_task_target; constantly pulls received
        messages from the queue and handles them

        assumes that message_receivers may be mutated, but the reference will not change

        can be canceled by draining the queue, if the queue is drainable, otherwise regular
        cancellation will work (but might interrupt a receiver, which is not ideal)
        """
        if (
            self.state.type == CRStateType.TORN_DOWN
            or self.state.type == CRStateType.ERRORED
        ):
            return
        receivers_dict = self.state.message_receivers
        errors: List[BaseException] = []
        while True:
            try:
                message = await queue.get()
            except QueueDrained:
                return
            if message.type == ReceivedMessageType.SMALL:
                await self._handle_received_small_message(
                    message, receivers_dict.values(), errors
                )
            elif message.type == ReceivedMessageType.LARGE:
                await self._handle_received_large_message(
                    message, receivers_dict.values(), errors
                )
            elif message.type == ReceivedMessageType.MISSED:
                received_missed.set()
            else:
                _assert_never(message)

            if errors:
                raise combine_multiple_exceptions(
                    "message handlers raised error", errors
                )

    async def _handle_received_message(
        self,
        message: Union[ReceivedSmallMessage, ReceivedLargeMessage],
        receivers: Iterable[PubSubDirectOnMessageWithCleanupReceiver],
        errors: List[BaseException],
    ) -> None:
        """normally received messages are handled in _handle_received_task_target,
        and if that is confirmed done(), this is called by _state_task_target
        with the final messages, such that this is never called concurrently
        with itself

        tells every receiver about the message, and if any of them raise an error,
        puts that error in the given list to be merged and raised at the appropriate
        point

        NOTE: the behavior that is targeted is that:
        - if one receiver is notified about a message, then every receiver is notified
        - receivers are notified one by one
        - once any message receiver errors, no future messages are sent to receivers and
          the connection is gracefully shutdown, then the error continues to bubble
          (the shutdown will cause the status handler to be informed of ABANDONED, which
          the caller will then see and teardown, which will have the error raised). the
          idea is that is perhaps the best stack trace experience: both the connection
          that was setup that errored and the receiver that errored will be included

        in some situations its recoverable if one message fails to process; in that case,
        just swallow errors inside the receiver (presumably, logging it where appropriate)
        """
        if message.type == ReceivedMessageType.SMALL:
            return await self._handle_received_small_message(message, receivers, errors)

        return await self._handle_received_large_message(message, receivers, errors)

    async def _handle_received_small_message(
        self,
        message: ReceivedSmallMessage,
        receivers: Iterable[PubSubDirectOnMessageWithCleanupReceiver],
        errors: List[BaseException],
    ) -> None:
        """informs receivers about a small message; see _handle_received_message for
        more information
        """
        for recv in list(receivers):
            try:
                await recv.on_message(
                    PubSubClientMessageWithCleanup(
                        topic=message.topic,
                        sha512=message.sha512,
                        data=io.BytesIO(message.data),
                        cleanup=_noop,
                    )
                )
            except BaseException as e:
                errors.append(e)

    async def _handle_received_large_message(
        self,
        message: ReceivedLargeMessage,
        receivers: Iterable[PubSubDirectOnMessageWithCleanupReceiver],
        errors: List[BaseException],
    ) -> None:
        """informs receivers about a large message; see _handle_received_message for
        more information
        """
        try:
            data_starts_at = message.stream.tell()
            for recv in list(receivers):
                try:
                    event = asyncio.Event()

                    async def _set_event() -> None:
                        event.set()

                    message.stream.seek(data_starts_at, os.SEEK_SET)
                    await recv.on_message(
                        PubSubClientMessageWithCleanup(
                            topic=message.topic,
                            sha512=message.sha512,
                            data=message.stream,
                            cleanup=_set_event,
                        )
                    )
                    await event.wait()
                except BaseException as e:
                    errors.append(e)
        finally:
            try:
                message.stream.close()
            except BaseException as e:
                errors.append(e)

    async def _check_errored(self) -> None:
        """Verifies we are in the SETUP state and the state task is still running.
        If closing and not retrying, waits for the state task to finish
        """
        if self.state.type == CRStateType.ERRORED:
            raise self.state.exception
        if self.state.type == CRStateType.TORN_DOWN:
            raise Exception("cannot use connection after teardown")

        assert self.state.type == CRStateType.SETUP, "call setup_connector first"
        if (
            self.state.ws_state.type == StateType.CLOSED
            or (
                self.state.ws_state.type == StateType.CLOSING
                and self.state.ws_state.retry.type
                == ClosingRetryInformationType.CANNOT_RETRY
            )
            or self.state.state_task.done()
        ):
            try:
                await self.state.state_task
                self.state = cast(CRState, self.state)  # tell mypy it may have changed
                if self.state.type == CRStateType.SETUP:
                    self.state = CRStateErrored(
                        type=CRStateType.ERRORED,
                        exception=Exception("state task finished unexpectedly"),
                    )
            except BaseException as e:
                self.state = cast(CRState, self.state)  # tell mypy it may have changed
                if self.state.type == CRStateType.SETUP:
                    self.state = CRStateErrored(type=CRStateType.ERRORED, exception=e)

            if self.state.type == CRStateType.ERRORED:
                raise self.state.exception

            assert self.state.type == CRStateType.TORN_DOWN
            raise Exception("cannot use connection after teardown")

    def _put_management_task(self, task: ManagementTask) -> None:
        """Puts a management task into the appropriate queue based on the current state
        to ensure it eventually gets performed

        MUST call `_check_errored` before calling this

        If it is a subscribe task then also ensures that the status receivers are
        told that the connection is LOST; this can be a little unintuitive, but the
        idea is that `on_connection_established` should be used if there was a period
        of time where messages may have been sent and not received, which is trivially
        true if the subscriber was previously not subscribed
        """
        assert self.state.type == CRStateType.SETUP, "_check_errored?"

        self.state.connection_lost_flag = (
            self.state.connection_lost_flag
            or task.type == ManagementTaskType.SUBSCRIBE_EXACT
            or task.type == ManagementTaskType.SUBSCRIBE_GLOB
        )

        if self.state.ws_state.type == StateType.OPEN:
            self.state.ws_state.management_tasks.put_nowait(task)
            return

        if self.state.ws_state.type == StateType.CLOSING:
            assert (
                self.state.ws_state.retry.type == ClosingRetryInformationType.WANT_RETRY
            )
            self.state.ws_state.retry.tasks.unsorted.put_nowait(task)
            return

        assert self.state.ws_state.type != StateType.CLOSED, "_check_errored?"
        self.state.ws_state.tasks.unsorted.put_nowait(task)

    async def setup_connector(self) -> None:
        await self._setup()

    async def teardown_connector(self) -> None:
        await self._teardown()

    async def subscribe_exact(self, /, *, topic: bytes) -> None:
        await self._check_errored()
        self._put_management_task(
            ManagementTaskSubscribeExact(
                type=ManagementTaskType.SUBSCRIBE_EXACT, topic=topic
            )
        )

    async def subscribe_glob(self, /, *, glob: str) -> None:
        await self._check_errored()
        self._put_management_task(
            ManagementTaskSubscribeGlob(
                type=ManagementTaskType.SUBSCRIBE_GLOB, glob=glob
            )
        )

    async def unsubscribe_exact(self, /, *, topic: bytes) -> None:
        await self._check_errored()
        self._put_management_task(
            ManagementTaskUnsubscribeExact(
                type=ManagementTaskType.UNSUBSCRIBE_EXACT, topic=topic
            )
        )

    async def unsubscribe_glob(self, /, *, glob: str) -> None:
        await self._check_errored()
        self._put_management_task(
            ManagementTaskUnsubscribeGlob(
                type=ManagementTaskType.UNSUBSCRIBE_GLOB, glob=glob
            )
        )

    def get_bulk(self) -> Optional[PubSubClientBulkSubscriptionConnector]:
        """Returns a bulk subscription connector if supported, otherwise None"""
        return None

    def prepare_notifier_trace(
        self, initializer: Literal[None], /
    ) -> ContextManager[PubSubClientTracingNotifyStart[Literal[None]]]:
        return _noop_notify_tracer()

    async def notify(
        self,
        /,
        *,
        topic: bytes,
        message: SyncStandardIO,
        length: int,
        message_sha512: bytes,
        tracer: Literal[None],
    ) -> PubSubNotifyResult:
        await self._check_errored()
        assert self.state.type == CRStateType.SETUP, "_check_errored?"

        state_queue: DrainableAsyncioQueue[InternalMessageState] = (
            DrainableAsyncioQueue(max_size=1)
        )
        callback = InternalMessageStateAndCallback(
            state=InternalMessageStateUnsent(
                type=InternalMessageStateType.UNSENT,
            ),
            callback=state_queue.put,
            task=None,
            queued=None,
        )

        identifier = secrets.token_bytes(4)
        msg: InternalMessage
        if (
            self.state.config.max_websocket_message_size is None
            or length < self.state.config.max_websocket_message_size
        ):
            msg = InternalSmallMessage(
                type=InternalMessageType.SMALL,
                identifier=identifier,
                topic=topic,
                data=read_exact(message, length),
                sha512=message_sha512,
                callback=callback,
            )
        else:
            message_start = message.tell()
            if message.tell() != 0:
                message = PositionedSyncStandardIO(
                    message,
                    start_idx=message_start,
                    end_idx=message_start + length,
                )
            msg = InternalLargeMessage(
                type=InternalMessageType.LARGE,
                identifier=identifier,
                topic=topic,
                stream=message,
                length=length,
                sha512=message_sha512,
                callback=callback,
            )

        if self.state.ws_state.type == StateType.OPEN:
            self.state.ws_state.unsent_notifications.put_nowait(msg)
        elif self.state.ws_state.type == StateType.CLOSING:
            assert (
                self.state.ws_state.retry.type == ClosingRetryInformationType.WANT_RETRY
            ), "_check_errored?"
            self.state.ws_state.retry.tasks.unsent_notifications.put_nowait(msg)
        elif self.state.ws_state.type == StateType.CLOSED:
            raise AssertionError("_check_errored?")
        else:
            self.state.ws_state.tasks.unsent_notifications.put_nowait(msg)

        async with state_queue:
            while True:
                state = await state_queue.get()
                if state.type == InternalMessageStateType.ACKNOWLEDGED:
                    return WsPubSubNotifyResult(state.notified)
                if state.type == InternalMessageStateType.DROPPED_UNSENT:
                    raise PubSubRequestConnectionAbandonedError(
                        "message was not sent before connection was lost"
                    )
                elif state.type == InternalMessageStateType.DROPPED_SENT:
                    raise PubSubRequestAmbiguousError(
                        "message was sent but not acknowledged before the connection was abandoned"
                    )

    @property
    def connection_status(self) -> PubSubClientConnectionStatus:
        if self.state.type == CRStateType.NOT_SETUP:
            return PubSubClientConnectionStatus.LOST
        if self.state.type != CRStateType.SETUP:
            return PubSubClientConnectionStatus.ABANDONED

        if (
            self.state.ws_state.type == StateType.OPEN
            # if there are no unsent management tasks and all the acks can be explained by sent
            # notifications, then all subscriptions/unsubscriptions have been ack'd
            and self.state.ws_state.management_tasks.qsize() == 0
            and len(self.state.ws_state.expected_acks)
            <= len(self.state.ws_state.sent_notifications)
            and not self.state.connection_lost_flag
        ):
            return PubSubClientConnectionStatus.OK

        if self.state.ws_state.type == StateType.CLOSED:
            return PubSubClientConnectionStatus.ABANDONED

        if (
            self.state.ws_state.type == StateType.CLOSING
            and self.state.ws_state.retry.type
            == ClosingRetryInformationType.CANNOT_RETRY
        ):
            return PubSubClientConnectionStatus.ABANDONED

        return PubSubClientConnectionStatus.LOST

    async def setup_receiver(self) -> None:
        await self._setup()

    async def teardown_receiver(self) -> None:
        await self._teardown()

    async def register_on_message(
        self, /, *, receiver: PubSubDirectOnMessageWithCleanupReceiver
    ) -> int:
        if self.state.type != CRStateType.NOT_SETUP:
            await self._check_errored()

        assert (
            self.state.type == CRStateType.SETUP
            or self.state.type == CRStateType.NOT_SETUP
        ), "_check_errored?"

        recv_id = self.state.receiver_counter
        self.state.receiver_counter += 1
        self.state.message_receivers[recv_id] = receiver
        return recv_id

    async def unregister_on_message(self, /, *, registration_id: int) -> None:
        if self.state.type != CRStateType.NOT_SETUP:
            await self._check_errored()

        assert (
            self.state.type == CRStateType.SETUP
            or self.state.type == CRStateType.NOT_SETUP
        ), "_check_errored?"

        self.state.message_receivers.pop(registration_id, None)

    async def register_status_handler(
        self, /, *, receiver: PubSubDirectConnectionStatusReceiver
    ) -> int:
        if self.state.type == CRStateType.NOT_SETUP:
            recv_id = self.state.receiver_counter
            self.state.receiver_counter += 1
            self.state.status_receivers[recv_id] = receiver
            return recv_id

        await self._check_errored()
        assert self.state.type == CRStateType.SETUP, "_check_errored?"
        recv_id = self.state.receiver_counter
        self.state.receiver_counter += 1
        self.state.new_status_receivers.put_nowait((recv_id, receiver))
        return recv_id

    async def unregister_status_handler(self, /, *, registration_id: int) -> None:
        if self.state.type == CRStateType.NOT_SETUP:
            self.state.status_receivers.pop(registration_id, None)
            return

        await self._check_errored()
        assert self.state.type == CRStateType.SETUP, "_check_errored?"
        self.state.removed_status_receivers.put_nowait(registration_id)


def WebsocketPubSubClient(config: WebsocketPubSubConfig) -> PubSubClient:
    """A constructor-like function that creates a PubSubClient that communicates
    over a websocket, handling retries and reconnections as necessary. It is
    typically important when using this to pass `on_receiving` when subscribing,
    as while reconnecting to broadcasters no messages will be received
    """

    async def setup() -> None:
        await config.setup_to_subscriber_auth()
        try:
            await config.setup_to_broadcaster_auth()
        except BaseException:
            await config.teardown_to_subscriber_auth()
            raise

    async def teardown() -> None:
        try:
            await config.teardown_to_broadcaster_auth()
        finally:
            await config.teardown_to_subscriber_auth()

    connector_receiver = WSPubSubConnectorReceiver(config)

    return PubSubClient(
        connector_receiver,
        connector_receiver,
        setup=setup,
        teardown=teardown,
    )


async def _noop() -> None:
    pass


def _assert_never(_: Never) -> NoReturn:
    raise AssertionError("unreachable")


if TYPE_CHECKING:
    _: Type[PubSubNotifyResult] = HttpPubSubNotifyResult
    __: Type[PubSubClientConnector] = WSPubSubConnectorReceiver
    ___: Type[PubSubClientReceiver] = WSPubSubConnectorReceiver



import io
import os
from types import TracebackType
from typing import TYPE_CHECKING, Any, Iterator, Optional, Type

from lonelypsc.types.sync_io import (
    SyncStandardIO,
    SyncStandardWithLengthIO,
)


class PositionedSyncStandardIO(io.IOBase):
    """Implements the SyncStandardWithLengthIO interface using the given part of the underlying
    stream. Also looks like an io.IOBase object to satisfy external callers, like aiohttp,
    which check against that rather than duck-typing
    """

    def __init__(self, stream: SyncStandardIO, start_idx: int, end_idx: int):
        assert 0 <= start_idx <= end_idx
        self.stream = stream
        self.start_idx = start_idx
        self.end_idx = end_idx
        self.index = min(stream.tell() - start_idx, end_idx - start_idx)

    def __iter__(self) -> Iterator[bytes]:
        return self

    def __next__(self) -> bytes:
        return self.read(io.DEFAULT_BUFFER_SIZE)

    def __enter__(self) -> "PositionedSyncStandardIO":
        raise NotImplementedError

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        raise NotImplementedError

    def close(self) -> None: ...

    def fileno(self) -> int:
        raise OSError("not implemented")

    def flush(self) -> None:
        raise OSError("not implemented")

    def isatty(self) -> bool:
        raise OSError("not implemented")

    def readable(self) -> bool:
        return True

    def readlines(self, hint: int = -1, /) -> list[bytes]:
        raise OSError("not implemented")

    def seekable(self) -> bool:
        return True

    def truncate(self, size: Optional[int] = None, /) -> int:
        raise OSError("not implemented")

    def writable(self) -> bool:
        return False

    def write(self, b: bytes) -> int:
        raise OSError("not implemented")

    def writelines(self, lines: Any, /) -> None:
        raise OSError("not implemented")

    def readline(self, size: Optional[int] = -1, /) -> bytes:
        raise OSError("not implemented")

    def __del__(self) -> None: ...

    @property
    def closed(self) -> bool:
        raise OSError("not implemented")

    def read(self, n: int) -> bytes:
        if n < 0:
            n = len(self) - self.index

        amount_to_read = min(len(self) - self.index, n)
        if amount_to_read > 0:
            actually_read = self.stream.read(amount_to_read)
            assert len(actually_read) <= amount_to_read
            self.index += len(actually_read)
            return actually_read
        return b""

    def tell(self) -> int:
        return self.index

    def seek(self, offset: int, whence: int = os.SEEK_SET) -> int:
        if whence == os.SEEK_SET:
            assert offset >= 0, "offset must be non-negative for SEEK_SET"
            target_notional_index = min(offset, len(self))
            target_real_index = target_notional_index + self.start_idx

            self.stream.seek(target_real_index, os.SEEK_SET)
            self.index = target_notional_index
            return self.index
        elif whence == os.SEEK_CUR:
            return self.seek(max(0, self.index + offset), os.SEEK_SET)
        elif whence == os.SEEK_END:
            return self.seek(max(0, len(self) + offset), os.SEEK_SET)
        raise OSError(f"unsupported whence: {whence}")

    def __len__(self) -> int:
        return self.end_idx - self.start_idx


if TYPE_CHECKING:
    _: Type[SyncStandardWithLengthIO] = PositionedSyncStandardIO


class PrefixedSyncStandardIO(io.IOBase):
    """Implements the SyncStandardWithLengthIO interface by acting like the entire prefix + entire child
    Also looks like an io.IOBase object to satisfy external callers, like aiohttp,
    which check against that rather than duck-typing
    """

    def __init__(
        self,
        prefix: SyncStandardWithLengthIO,
        child: SyncStandardWithLengthIO,
    ):
        self.prefix = prefix
        self.child = child
        self.index = 0
        self.prefix.seek(0)
        if len(self.prefix) == 0:
            self.child.seek(0)

    def __iter__(self) -> Iterator[bytes]:
        return self

    def __next__(self) -> bytes:
        return self.read(io.DEFAULT_BUFFER_SIZE)

    def __enter__(self) -> "PrefixedSyncStandardIO":
        raise NotImplementedError

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        raise NotImplementedError

    def close(self) -> None: ...

    def fileno(self) -> int:
        raise OSError("not implemented")

    def flush(self) -> None:
        raise OSError("not implemented")

    def isatty(self) -> bool:
        raise OSError("not implemented")

    def readable(self) -> bool:
        return True

    def readlines(self, hint: int = -1, /) -> list[bytes]:
        raise OSError("not implemented")

    def seekable(self) -> bool:
        return True

    def truncate(self, size: Optional[int] = None, /) -> int:
        raise OSError("not implemented")

    def writable(self) -> bool:
        return False

    def write(self, b: bytes) -> int:
        raise OSError("not implemented")

    def writelines(self, lines: Any, /) -> None:
        raise OSError("not implemented")

    def readline(self, size: Optional[int] = -1, /) -> bytes:
        raise OSError("not implemented")

    def __del__(self) -> None: ...

    @property
    def closed(self) -> bool:
        raise OSError("not implemented")

    def read(self, n: int) -> bytes:
        if n < 0:
            n = len(self) - self.index

        result = b""

        amount_to_read_from_prefix = min(len(self.prefix) - self.index, n)
        if amount_to_read_from_prefix > 0:
            actually_read = self.prefix.read(amount_to_read_from_prefix)
            assert len(actually_read) <= amount_to_read_from_prefix
            self.index += len(actually_read)
            n -= len(actually_read)
            if len(actually_read) < amount_to_read_from_prefix:
                return actually_read
            result = actually_read

        amount_to_read = min(len(self) - self.index, n)
        if amount_to_read > 0:
            actually_read = self.child.read(amount_to_read)
            assert len(actually_read) <= amount_to_read
            self.index += len(actually_read)
            result += actually_read

        return result

    def tell(self) -> int:
        return self.index

    def seek(self, offset: int, whence: int = os.SEEK_SET) -> int:
        if whence == os.SEEK_SET:
            assert offset >= 0, "offset must be non-negative for SEEK_SET"
            if offset < len(self.prefix):
                self.prefix.seek(offset)
                self.index = offset
                return offset

            if offset >= len(self):
                self.index = len(self)
                return self.index

            offset_in_child = offset - len(self.prefix)
            self.child.seek(offset_in_child)
            self.index = offset
            return self.index
        elif whence == os.SEEK_CUR:
            return self.seek(self.index + offset, os.SEEK_SET)
        elif whence == os.SEEK_END:
            return self.seek(len(self) + offset, os.SEEK_SET)
        raise OSError(f"unsupported whence: {whence}")

    def __len__(self) -> int:
        return len(self.prefix) + len(self.child)


if TYPE_CHECKING:
    __: Type[SyncStandardWithLengthIO] = PrefixedSyncStandardIO



import random
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    TypedDict,
    Union,
    cast,
)


class SimplePubSubBroadcasterConfig(TypedDict):
    """Indicates how to connect to a broadcaster. We leave this type open for
    expansion in the future (compared to a string) in case we want to add e.g.
    optional priority levels
    """

    host: str
    """The host of the broadcaster, e.g., `http://localhost:3003`. Must include
    the schema and port (if not default) and may include a path (if prefixing the
    standard paths), such that f'{host}/v1/subscribe/exact' is the full path to 
    subscribe exact endpoint.

    For websocket clients, the schema must be a ws or wss. For http clients,
    the schema must be http or https.
    """


class PrioritizedPubSubBroadcasterConfig(SimplePubSubBroadcasterConfig):
    """Indicates how to connect to a broadcaster with a priority level. This is
    primarily recommended for websocket connections
    """

    priority: int
    """The priority level of the broadcaster. Lower values are higher priority.
    Ties are broken randomly. All specific priority values are considered higher
    priority (will be used first) than an unset priority value.

    Priority is most important when using websockets as a bad choice of host has
    much longer lasting effects compared to http, and websockets are generally
    used where there is more volume and thus the latency is more important
    """


PubSubBroadcasterConfig = Union[
    SimplePubSubBroadcasterConfig, PrioritizedPubSubBroadcasterConfig
]


class _BroadcasterShufflerIterator:
    """The iterator produced by BroadcastersShuffler.

    There are two design goals:
    1. Correctness; this should be a fair shuffling of the broadcasters,
       and should respect priority
    2. Efficiency in happy path; this should have essentially no overhead when
       it's used with only a single next() call then discarded, as is expected
       most of the time
    3. Reasonable efficiency in unhappy path; this should be able to handle
       iterating through the entire list in no worse than O(n (log n)^2) time
    """

    def __init__(self, parent: "BroadcastersShuffler") -> None:
        self.parent = parent
        self._level_start: int = 0
        self._level_end: Optional[int] = None
        self._level_first_used_idx: Optional[int] = None
        self._level_remaining_idxs: Optional[List[int]] = None

    def __next__(self) -> PubSubBroadcasterConfig:
        if self._level_start >= len(self.parent.sorted_broadcasters):
            raise StopIteration

        if self._level_end is None:
            # typically there will be either be very few options (e.g., 1) at
            # this level (when using priority to prefer connecting to the nearest
            # instance) or the entire remaining list will be at this level
            #
            # example 1: using multiple availability zones within the same region
            # [{"priority": 0, "host": "same az"}, {"host": "other az"}, {"host": "other az"}, ...]
            #
            # example 2: using multiple availablity zones and multiple regions
            # [{"priority": 0, "host": "same region, same az"}, {"priority": 1, "host": "same region, other az"}, {"host": "other region"}, {"host": "other region"}, ...]

            # over all next() calls this never repeats the same index, so requires O(n)
            # time
            level = self.parent.sorted_broadcasters[self._level_start].get("priority")
            if self.parent.sorted_broadcasters[-1].get("priority") == level:
                self._level_end = len(self.parent.sorted_broadcasters)
            else:
                self._level_end = self._level_start + 1
                while (
                    self._level_end < len(self.parent.sorted_broadcasters)
                    and self.parent.sorted_broadcasters[self._level_end].get("priority")
                    == level
                ):
                    self._level_end += 1

        if self._level_first_used_idx is None:
            # most common case: first pluck from this level is very easy
            # if we don't end up iterating through the rest, we can avoid
            # building the remaining list
            # happens at most once per level, which is at most once per index.
            # random.choice is constant time in this case, so requires at most
            # O(n) time
            choice = random.choice(range(self._level_start, self._level_end))

            if self._level_start == self._level_end - 1:
                self._level_start += 1
                self._level_end = None
            else:
                self._level_first_used_idx = choice

            return self.parent.sorted_broadcasters[choice]

        if self._level_remaining_idxs is None:
            # second pluck from this level
            # this never repeats the same index and is equally performant
            # regardless of the number of levels as building the indexes +
            # random.shuffle takes O(n) time
            # this takes O(n) time
            self._level_remaining_idxs = list(
                i
                for i in range(self._level_start, self._level_end)
                if i != self._level_first_used_idx
            )
            random.shuffle(self._level_remaining_idxs)

        # constant time pop up to n-1 times; requires O(n) time
        choice = self._level_remaining_idxs.pop()
        if not self._level_remaining_idxs:
            self._level_start = self._level_end
            self._level_end = None
            self._level_first_used_idx = None
            self._level_remaining_idxs = None

        return self.parent.sorted_broadcasters[choice]

    def __iter__(self) -> "_BroadcasterShufflerIterator":
        return self


class BroadcastersShuffler:
    """An object that can precompute whatever internal data structure is
    appropriate such that it can quickly build new iterators for the broadcasters
    that match the desired semantics

    Usage:

    ```python
    from lonelypsc.config.config import BroadcastersShuffler

    shuffler = BroadcastersShuffler(
        [
            {"host": "http://localhost:3003"},
            {"host": "http://localhost:3004"},
            {"host": "http://localhost:3005", "priority": 0},
        ]
    )
    for broadcaster in shuffler:
        print(broadcaster)
    ```
    """

    def __init__(self, broadcasters: List[PubSubBroadcasterConfig]) -> None:
        self.sorted_broadcasters = sorted(
            broadcasters,
            key=lambda x: cast(Dict[str, int], x).get("priority", float("+inf")),
        )

    def __iter__(self) -> _BroadcasterShufflerIterator:
        return _BroadcasterShufflerIterator(self)


if TYPE_CHECKING:
    _: Type[Iterator[PubSubBroadcasterConfig]] = _BroadcasterShufflerIterator
    __: Type[Iterable[PubSubBroadcasterConfig]] = BroadcastersShuffler
    ___: PubSubBroadcasterConfig = {"host": "http://localhost:3003"}
    ____: PubSubBroadcasterConfig = {
        "host": "http://localhost:3003",
        "priority": 0,
    }



import asyncio
import time
from typing import TYPE_CHECKING

import aiohttp

from lonelypsc.client import PubSubCancelRequested, PubSubIrrecoverableError
from lonelypsc.ws.check_result import (
    CheckResult,
    CheckStateChangerResult,
    CheckStateChangerResultContinue,
    CheckStateChangerResultDone,
)
from lonelypsc.ws.handle_connection_failure import (
    cleanup_tasks_and_raise,
)
from lonelypsc.ws.handlers.protocol import StateHandler
from lonelypsc.ws.handlers.util.state_specific_cleanup import handle_via_composition
from lonelypsc.ws.internal_callbacks import sweep_internal_message
from lonelypsc.ws.state import (
    State,
    StateConnecting,
    StateType,
    StateWaitingRetry,
)
from lonelypsc.ws.websocket_connect_task import make_websocket_connect_task


async def handle_waiting_retry(state: State) -> State:
    """
    Lets the backgrounded tasks finish, waits for the retry delay to pass, then moves
    to the CONNECTING state
    """
    assert state.type == StateType.WAITING_RETRY
    return await handle_via_composition(
        state=state, core=_core, recover=_recover, shutdown=_shutdown
    )


async def _core(state: StateWaitingRetry) -> State:
    """Happy path to move from WAITING_RETRY to CONNECTING, raises an exception to
    request recover/shutdown
    """
    if (result := await _check_canceled(state)).type == CheckResult.RESTART:
        return result.state

    if (result := await _check_timeout(state)).type == CheckResult.RESTART:
        return result.state

    _sweep_backgrounded(state)
    _sweep_resending(state)
    await _wait_something_changed(state)
    return state


async def _wait_something_changed(state: StateWaitingRetry) -> None:
    """Waits for something to need to be done about the given state"""
    wait_cancel = asyncio.create_task(state.cancel_requested.wait())
    wait_timeout = asyncio.create_task(asyncio.sleep(state.retry_at - time.time()))
    await asyncio.wait(
        [
            wait_cancel,
            wait_timeout,
            *state.backgrounded,
            *[
                msg.callback.task
                for msg in state.tasks.resending_notifications
                if msg.callback.task is not None
            ],
        ],
        return_when=asyncio.FIRST_COMPLETED,
    )
    wait_cancel.cancel()
    wait_timeout.cancel()


async def _check_canceled(state: StateWaitingRetry) -> CheckStateChangerResult:
    if not state.cancel_requested.is_set():
        return CheckStateChangerResultContinue(type=CheckResult.CONTINUE)

    raise PubSubCancelRequested()


def _sweep_backgrounded(state: StateWaitingRetry) -> None:
    """Cleans out done background tasks, raising an irrecoverable error if any failed"""
    new_backgrounded = set()

    for bknd in state.backgrounded:
        if not bknd.done():
            new_backgrounded.add(bknd)
            continue

        if bknd.exception() is not None:
            raise PubSubIrrecoverableError("background task failed")

    state.backgrounded = new_backgrounded


def _sweep_resending(state: StateWaitingRetry) -> None:
    """Progresses resending notification state callbacks, raising an error if any failed"""
    for msg in state.tasks.resending_notifications:
        sweep_internal_message(msg)


async def _check_timeout(state: StateWaitingRetry) -> CheckStateChangerResult:
    """Checks if the retry delay has passed"""
    if time.time() < state.retry_at:
        return CheckStateChangerResultContinue(type=CheckResult.CONTINUE)

    # this should be the first next on this iterator
    try:
        broadcaster = next(state.retry.iterator)
    except StopIteration:
        raise PubSubIrrecoverableError("no broadcasters to try")

    client_session = aiohttp.ClientSession()
    return CheckStateChangerResultDone(
        type=CheckResult.RESTART,
        state=StateConnecting(
            type=StateType.CONNECTING,
            config=state.config,
            client_session=client_session,
            websocket_task=make_websocket_connect_task(
                state.config, broadcaster, client_session
            ),
            cancel_requested=state.cancel_requested,
            broadcaster=broadcaster,
            retry=state.retry,
            tasks=state.tasks,
            backgrounded=state.backgrounded,
        ),
    )


async def _recover(state: StateWaitingRetry, /, *, cause: BaseException) -> State:
    """Recovery is not possible for WAITING_RETRY, so re-raises, marking the
    cause as irrecoverable
    """
    raise PubSubIrrecoverableError("recovery not possible") from cause


async def _shutdown(state: StateWaitingRetry, /, *, cause: BaseException) -> State:
    """Move the state machine towards the CLOSED state by cleaning up and bubbling"""
    await cleanup_tasks_and_raise(
        state.tasks, state.backgrounded, "failed to retry", cause
    )


if TYPE_CHECKING:
    _: StateHandler = handle_waiting_retry



from lonelypsp.auth.config import AuthResult

from lonelypsc.client import PubSubError
from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.handlers.open.handle_authorized_receive import (
    handle_authorized_receive,
)
from lonelypsc.ws.state import ReceivingState, StateOpen


def check_receiving_authorizing(state: StateOpen) -> CheckResult:
    """
    Tries to move from receiving state `AUTHORIZING` to either WAITING_COMPRESSOR or None

    Raises an error if the authorization result is not "ok"
    """
    if state.receiving is None or state.receiving.type != ReceivingState.AUTHORIZING:
        return CheckResult.CONTINUE

    if not state.receiving.authorization_task.done():
        return CheckResult.CONTINUE

    result = state.receiving.authorization_task.result()
    if result != AuthResult.OK:
        raise PubSubError(f"authorization failed: {result}")

    handle_authorized_receive(state, state.receiving.first, state.receiving.body)
    return CheckResult.RESTART



import asyncio
import hashlib
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Iterator, List, Literal, Optional, Protocol, Set, Union

import aiohttp
from aiohttp import ClientSession, ClientWebSocketResponse
from lonelypsp.auth.config import AuthResult
from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import SubscriberToBroadcasterStatefulMessageType
from lonelypsp.stateful.messages.confirm_notify import B2S_ConfirmNotify
from lonelypsp.stateful.messages.confirm_subscribe import (
    B2S_ConfirmSubscribeExact,
    B2S_ConfirmSubscribeGlob,
)
from lonelypsp.stateful.messages.confirm_unsubscribe import (
    B2S_ConfirmUnsubscribeExact,
    B2S_ConfirmUnsubscribeGlob,
)
from lonelypsp.stateful.messages.continue_notify import B2S_ContinueNotify
from lonelypsp.stateful.messages.missed import B2S_Missed
from lonelypsp.stateful.messages.receive_stream import (
    B2S_ReceiveStreamStartCompressed,
    B2S_ReceiveStreamStartUncompressed,
)
from lonelypsp.util.bounded_deque import BoundedDeque
from lonelypsp.util.drainable_asyncio_queue import DrainableAsyncioQueue

from lonelypsc.config.config import BroadcastersShuffler, PubSubBroadcasterConfig
from lonelypsc.config.ws_config import WebsocketPubSubConfig
from lonelypsc.types.sync_io import SyncIOBaseLikeIO, SyncStandardIO
from lonelypsc.types.websocket_message import WSMessage
from lonelypsc.ws.compressor import CompressorStore


class StateType(Enum):
    """Discriminator value for the state the websocket is in"""

    CONNECTING = auto()
    """The necessary websocket/tcp handshakes for communicating with a specific
    broadcaster are in progress. This is the initial state, and the state that
    the websocket will return to if the connection is closed unexpectedly and there
    are still retries available
    """

    CONFIGURING = auto()
    """A websocket connection is open to a specific broadcaster and either the configure
    message is being sent or the subscriber is waiting to receive the configure confirmation
    message from the broadcaster
    """

    OPEN = auto()
    """The standard state where a websocket connection is open with the broadcaster
    and the configuration handshake is already complete.
    """

    WAITING_RETRY = auto()
    """No websocket connection is open or opening but the subscriber plans to try 
    connecting to broadcasters again after a retry period
    """

    CLOSING = auto()
    """Trying to close the websocket normally, then possibly raising an exception"""

    CLOSED = auto()
    """There is no websocket connection open and no plans to retry connecting to
    broadcasters, either because the subscriber chose to disconnect or because
    all broadcasters and retries have been exhausted
    """


@dataclass
class RetryInformation:
    """Information required to progress through other broadcasters and
    retries after a failure to connect to a broadcaster
    """

    shuffler: BroadcastersShuffler
    """The shuffler that is being used to produce iterators of broadcasters"""

    iteration: int
    """Starts at 0 indicating that there are some broadcasters that have
    not been tried at all, then after all broadcasters have been attempted
    increases to 1, then 2, etc.
    """

    iterator: Iterator[PubSubBroadcasterConfig]
    """The iterator for this iteration of the shuffler"""


class OpenRetryInformationType(Enum):
    """discriminator value for OpenRetryInformation.type"""

    STABLE = auto()
    """the connection has been stable long enough that if disconnected the
    subscriber will restart the retry process from the beginning
    """

    TENTATIVE = auto()
    """the connection was recently opened, so if disconnected the subscriber
    will continue the retry process from where it left off
    """


@fast_dataclass
class OpenRetryInformationStable:
    type: Literal[OpenRetryInformationType.STABLE]
    """discriminator value"""


@fast_dataclass
class OpenRetryInformationTentative:
    type: Literal[OpenRetryInformationType.TENTATIVE]
    """discriminator value"""

    stable_at: float
    """if the connection is still live at this time in fractional seconds since
    the unix epoch, then the connection should be considered stable
    """

    continuation: RetryInformation
    """how to continue the retry process if disconnected"""


OpenRetryInformation = Union[OpenRetryInformationStable, OpenRetryInformationTentative]


class ClosingRetryInformationType(Enum):
    """discriminator value for ClosingRetryInformation.type"""

    CANNOT_RETRY = auto()
    """There is an exception that cannot be caught; it must be raised
    after closing the websocket
    """

    WANT_RETRY = auto()
    """the websocket is closing unexpectedly and there will be an attempt to
    re-establish a connection. move to WAITING_RETRY state, move to the
    CONNECTING state, or raise that all retries have been exhausted based on the
    result of next() on the iterator
    """


@fast_dataclass
class ClosingRetryInformationCannotRetry:
    type: Literal[ClosingRetryInformationType.CANNOT_RETRY]
    """discriminator value"""

    tasks: Optional["TasksOnceOpen"]
    """the tasks that need to be cleaned up, if any"""

    exception: BaseException
    """the exception to raise once the websocket is closed"""


@fast_dataclass
class ClosingRetryInformationWantRetry:
    type: Literal[ClosingRetryInformationType.WANT_RETRY]
    """discriminator value"""

    retry: RetryInformation
    """how to continue the retry process"""

    tasks: "TasksOnceOpen"
    """the tasks that need to be performed after re-establishing the
    connection
    """

    exception: BaseException
    """context for the exception that indicated the connection needed to
    be closed
    """


ClosingRetryInformation = Union[
    ClosingRetryInformationCannotRetry, ClosingRetryInformationWantRetry
]


class InternalMessageType(Enum):
    """Discriminator value for the type of message that is being sent internally"""

    SMALL = auto()
    """The message is small enough to hold in memory"""

    LARGE = auto()
    """The message can be read in parts via a stream"""


class InternalMessageStateType(Enum):
    """Describes what state an internal message can be in"""

    UNSENT = auto()
    """
    - the subscriber HAS NOT sent the message
    - the broadcaster HAS NOT processed the message
    - the state WILL change again
    - read() MAY be called on the stream (for large messages)
    """

    SENT = auto()
    """
    - the subscriber MAY HAVE sent the message ANY NUMBER OF TIMES
    - the broadcaster MAY HAVE processed the message ANY NUMBER OF TIMES
    - the state WILL change again
    - read() WILL NOT be called on the stream (for large messages)
    """

    RESENDING = auto()
    """
    - the subscriber MAY HAVE sent the message ANY NUMBER OF TIMES
    - the broadcaster MAY HAVE processed the message ANY NUMBER OF TIMES
    - the state WILL change again
    - read() MAY be called on the stream (for large messages)
    """

    ACKNOWLEDGED = auto()
    """
    - the subscriber HAS sent the message AT LEAST ONCE
    - the broadcaster HAS processed the message AT LEAST ONCE
    - the state WILL NOT change again
    - read() WILL NOT be called on the stream (for large messages)
    """

    DROPPED_UNSENT = auto()
    """
    - the subscriber MAY HAVE sent the message ANY NUMBER OF TIMES
    - the broadcaster HAS NOT processed the message
    - the state WILL NOT change again
    - read() WILL NOT be called on the stream (for large messages)

    note: the subscriber MAY move the message from `SENT` to `DROPPED_UNSENT`
    if the broadcaster explicitly refuses the message, though this is not
    currently implemented. this is because its more important for most callers
    to know if the broadcaster may have processed the message vs the message 
    was sent
    """

    DROPPED_SENT = auto()
    """
    - the subscriber MAY HAVE sent the message ANY NUMBER OF TIMES
    - the broadcaster MAY HAVE processed the message ANY NUMBER OF TIMES
    - the state WILL NOT change again
    - read() WILL NOT be called on the stream (for large messages)
    """


@fast_dataclass
class InternalMessageStateUnsent:
    type: Literal[InternalMessageStateType.UNSENT]
    """discriminator value"""


@fast_dataclass
class InternalMessageStateSent:
    type: Literal[InternalMessageStateType.SENT]
    """discriminator value"""


@fast_dataclass
class InternalMessageStateResending:
    type: Literal[InternalMessageStateType.RESENDING]
    """discriminator value"""


@fast_dataclass
class InternalMessageStateAcknowledged:
    type: Literal[InternalMessageStateType.ACKNOWLEDGED]
    """discriminator value"""

    notified: int
    """a lower bound for the number of unique subscribers that received the
    message
    """


@fast_dataclass
class InternalMessageStateDroppedUnsent:
    type: Literal[InternalMessageStateType.DROPPED_UNSENT]
    """discriminator value"""


@fast_dataclass
class InternalMessageStateDroppedSent:
    type: Literal[InternalMessageStateType.DROPPED_SENT]
    """discriminator value"""


InternalMessageState = Union[
    InternalMessageStateUnsent,
    InternalMessageStateSent,
    InternalMessageStateResending,
    InternalMessageStateAcknowledged,
    InternalMessageStateDroppedUnsent,
    InternalMessageStateDroppedSent,
]


class InternalMessageStateCallback(Protocol):
    """Describes an function that is called when a message changes state
    (e.g., from unsent to sent, from sent to acknowledged, from any to dropped,
    etc). See the documentation for the `InternalMessageStateType` enum for
    details on what each state means

    It is guarranteed that within a single message this callback is not invoked
    concurrently, so e.g. the following implementation will always produce a
    valid ordering of internal message states:

    ```python
    async def my_callback(state: InternalMessageState) -> None:
        await asyncio.sleep(random.random())  # simulates network io
        print(f"state: {state.type}")  # stdout ordered correctly
    ```

    Note that for the most part a valid ordering just means that a final state
    will be printed last, i.e., `UNSENT -> SENT -> ACKNOWLEDGED`, or possibly
    `UNSENT -> ACKNOWLEDGED` (skipping an intermediate state), but definitely
    *not* `UNSENT -> ACKNOWLEDGED -> SENT`
    """

    async def __call__(self, state: InternalMessageState, /) -> None:
        pass


@dataclass
class InternalMessageStateAndCallback:
    """Keeps track of what state a callback was last scheduled for on the
    event loop, the running task for the callback (if any), and the the queued
    state for the callback

    By not weaving multiple calls to the same callback together it is easier
    to reason about recovery when using this library; the alternative is that
    this library only promises the order that the callbacks are scheduled is
    consistent, but if e.g. logging is async this easily leads to subtle bugs
    that are hard to distinguish from errors within the library
    """

    state: InternalMessageState
    """the last state passed to the callback; if task is set, this is the
    state sent to the callback that the task is tracking
    """

    callback: InternalMessageStateCallback
    """the actual callback function"""

    task: Optional[asyncio.Task[None]]
    """the task that is running the callback right now, if any. the state can only
    be changed if this task is None or from within this task
    """

    queued: Optional[
        Union[
            InternalMessageStateUnsent,
            InternalMessageStateSent,
            InternalMessageStateResending,
        ]
    ]
    """the most recent intermediate state for the callback, or None if the
    callback knows about the most recent intermediate state.

    NOTE: the final state is handled differently and, at that point there are
    assumed to be no lingering references to this object
    """


@fast_dataclass
class InternalSmallMessage:
    """
    A message to be sent that is entirely in memory. although this could be
    converted to an InternalLargeMessage via wrapping the data with BytesIO, if
    the whole message is in memory it is often possible to be more efficient
    than generic stream processing, so this distinction is still useful
    """

    type: Literal[InternalMessageType.SMALL]
    """discriminator value"""

    identifier: bytes
    """the arbitrary, unique identifier the subscriber assigned to this message"""

    topic: bytes
    """the topic the message was sent to"""

    data: bytes
    """the uncompressed message data"""

    sha512: bytes
    """the trusted 64-byte hash of the data"""

    callback: InternalMessageStateAndCallback
    """the state of the callback and the callback for this message"""


@fast_dataclass
class InternalLargeMessage:
    """A message that can be read in parts via a stream"""

    type: Literal[InternalMessageType.LARGE]
    """discriminator value"""

    identifier: bytes
    """the arbitrary, unique identifier the subscriber assigned to this message"""

    topic: bytes
    """the topic the message was sent to"""

    stream: SyncStandardIO
    """the readable, seekable, tellable stream that the message data is read from

    guarranteed to have data starting at tell() == 0

    no read() calls that would read past the indicated length from the stream will 
    be made

    this stream is never closed by the state machine, but the caller is alerted
    to if its functions will be called via the callback field (for
    example, if this stream can be reopened the caller could e.g. close in SENT
    then reopen in RESENDING).
    """

    length: int
    """the total length that can be read from the stream; the stream may over-read
    if read(n) is called with an n too large, which would be an error in lonelypsc
    not the caller
    """

    sha512: bytes
    """the trusted 64-byte SHA512 hash of the data"""

    callback: InternalMessageStateAndCallback
    """the state of the callback and the callback for this message"""


InternalMessage = Union[InternalSmallMessage, InternalLargeMessage]


class ManagementTaskType(Enum):
    """discriminator value for `ManagementTask.type`"""

    SUBSCRIBE_EXACT = auto()
    """need to subscribe to an exact topic"""

    SUBSCRIBE_GLOB = auto()
    """need to subscribe to a glob pattern"""

    UNSUBSCRIBE_EXACT = auto()
    """need to unsubscribe from an exact topic"""

    UNSUBSCRIBE_GLOB = auto()
    """need to unsubscribe from a glob pattern"""

    GRACEFUL_DISCONNECT = auto()
    """graceful disconnect requested"""


@fast_dataclass
class ManagementTaskSubscribeExact:
    type: Literal[ManagementTaskType.SUBSCRIBE_EXACT]
    """discriminator value"""

    topic: bytes
    """the topic to subscribe to"""


@fast_dataclass
class ManagementTaskSubscribeGlob:
    type: Literal[ManagementTaskType.SUBSCRIBE_GLOB]
    """discriminator value"""

    glob: str
    """the glob pattern to subscribe to"""


@fast_dataclass
class ManagementTaskUnsubscribeExact:
    type: Literal[ManagementTaskType.UNSUBSCRIBE_EXACT]
    """discriminator value"""

    topic: bytes
    """the topic to unsubscribe from"""


@fast_dataclass
class ManagementTaskUnsubscribeGlob:
    type: Literal[ManagementTaskType.UNSUBSCRIBE_GLOB]
    """discriminator value"""

    glob: str
    """the glob pattern to unsubscribe from"""


ManagementTask = Union[
    ManagementTaskSubscribeExact,
    ManagementTaskSubscribeGlob,
    ManagementTaskUnsubscribeExact,
    ManagementTaskUnsubscribeGlob,
]


@fast_dataclass
class TasksOnceOpen:
    """When not in the OPEN state the client can still receive requests to
    perform operations (e.g., subscribe, notify). This object keeps track
    of those operations that need to be performed until the OPEN state, where
    they are transformed into a different form for actually being sent across
    the websocket
    """

    exact_subscriptions: Set[bytes]
    """The topics which should be subscribed to"""

    glob_subscriptions: Set[str]
    """The glob patterns which should be subscribed to"""

    unsorted: DrainableAsyncioQueue[ManagementTask]
    """management tasks which have been sent from other asyncio coroutines and not applied yet"""

    unsent_notifications: DrainableAsyncioQueue[InternalMessage]
    """The unsent messages that should be sent to the broadcaster via NOTIFY / NOTIFY STREAM."""

    resending_notifications: List[InternalMessage]
    """The resending messages that should be sent to the broadcaster via NOTIFY / NOTIFY STREAM"""


Acknowledgement = Union[
    B2S_ConfirmSubscribeExact,
    B2S_ConfirmSubscribeGlob,
    B2S_ConfirmUnsubscribeExact,
    B2S_ConfirmUnsubscribeGlob,
    B2S_ContinueNotify,
    B2S_ConfirmNotify,
]


class ReceivedMessageType(Enum):
    """Discriminator value for `ReceivedMessage.type`"""

    SMALL = auto()
    """The message is entirely in memory"""

    MISSED = auto()
    """The actual message was never sent"""

    LARGE = auto()
    """The message is in a stream"""


@fast_dataclass
class ReceivedSmallMessage:
    """A received message which is entirely in memory"""

    type: Literal[ReceivedMessageType.SMALL]
    """discriminator value"""

    topic: bytes
    """the topic the message was sent to"""

    data: bytes
    """the uncompressed message data"""

    sha512: bytes
    """the trusted 64-byte hash of the data"""


@fast_dataclass
class ReceivedMissedMessage:
    """indicates that the broadcaster may not have sent a message it should have"""

    type: Literal[ReceivedMessageType.MISSED]
    """discriminator value"""

    topic: bytes
    """the topic the missed message may have been on"""


@fast_dataclass
class ReceivedLargeMessage:
    """A received message which is not entirely in memory; closing the
    stream will delete the data. must close the stream once it is
    consumed
    """

    type: Literal[ReceivedMessageType.LARGE]
    """discriminator value"""

    topic: bytes
    """the topic the message was sent to"""

    stream: SyncIOBaseLikeIO
    """the readable, seekable, tellable, closeable stream that the message data can
    be read from
    """

    sha512: bytes
    """the trusted 64-byte hash of the data"""


ReceivedMessage = Union[
    ReceivedSmallMessage, ReceivedMissedMessage, ReceivedLargeMessage
]


class ReceivingState(Enum):
    """Discriminator value for `StateOpen.receiving`"""

    INCOMPLETE = auto()
    """Waiting for the rest of the message to come in"""

    AUTHORIZING_MISSED = auto()
    """Waiting for the authorization task to complete on a missed message"""

    AUTHORIZING = auto()
    """Waiting for the authorization task to complete"""

    WAITING_COMPRESSOR = auto()
    """Waiting for the compressor to be ready"""

    DECOMPRESSING = auto()
    """Decompressing the message"""


@dataclass
class ReceivingIncomplete:
    """
    The subscriber has received part of a message via a RECEIVE_STREAM and is
    waiting for more to come in. The subscriber is hashing the message as it
    comes in, and may have not finished verifying the provided authorization.
    The subscriber hasn't confirmed the hash from the first message yet, since
    that requires the entire message to be received, so the authorization check
    is tentative anyway
    """

    type: Literal[ReceivingState.INCOMPLETE]
    """disciminator value"""

    first: Union[B2S_ReceiveStreamStartUncompressed, B2S_ReceiveStreamStartCompressed]
    """The first stream message with this id, with the payload stripped out"""

    part_id: int
    """The last part id that the subscriber received"""

    body_hasher: "hashlib._Hash"
    """the hash object that is producing the sha512 hash of the body as it comes in"""

    body: SyncIOBaseLikeIO
    """a writable, seekable, tellable, closeable file-like object where the
    subscriber is storing the potentially compressed body of the message as it
    comes in. closing this file will delete the data
    """

    authorization_task: Optional[asyncio.Task[AuthResult]]
    """the task checking if the authorization on the first message is valid or None
    if the task completed, was already checked, and was "ok"
    """


@fast_dataclass
class ReceivingAuthorizingMissed:
    """The subscriber has received a MISSED message and is waiting to verify that
    the authorization is valid before proceeding
    """

    type: Literal[ReceivingState.AUTHORIZING_MISSED]
    """disciminator value"""

    message: B2S_Missed
    """the message that was received"""

    authorization_task: asyncio.Task[AuthResult]
    """the task the subscriber is waiting on to finish checking the messages authorization"""


@fast_dataclass
class ReceivingAuthorizing:
    """The subscriber has received the entire RECEIVE via RECEIVE_STREAM calls, verified that the
    hash matches what was provided, and are waiting for the authorization task to complete
    before proceeding

    the message has already been ack'd (or at least the ack has been queued) at this point
    """

    type: Literal[ReceivingState.AUTHORIZING]
    """disciminator value"""

    first: Union[B2S_ReceiveStreamStartUncompressed, B2S_ReceiveStreamStartCompressed]
    """The first stream message with this id, with the payload stripped out"""

    body: SyncIOBaseLikeIO
    """a readable, seekable, tellable, closeable file-like object where the
    subscriber stored the potentially compressed body of the message as it came
    in. closing this file will delete the data
    """

    authorization_task: asyncio.Task[AuthResult]
    """the task the subscriber is waiting on to finish checking the messages authorization"""


@fast_dataclass
class ReceivingWaitingCompressor:
    """
    The subscriber has received the entire message via one or more
    RECEIVE_STREAM calls, verified that the hash matches what was provided,
    verified the authorization provided, and is waiting for the compressor to be
    ready
    """

    type: Literal[ReceivingState.WAITING_COMPRESSOR]
    """disciminator value"""

    first: B2S_ReceiveStreamStartCompressed
    """The first stream message with this id, with the payload stripped out"""

    compressed_body: SyncIOBaseLikeIO
    """a readable, seekable, tellable, closeable file-like object where the subscriber
    stored the compressed body of the message as it came in. closing this file will delete
    the data
    """


@fast_dataclass
class ReceivingDecompressing:
    """
    The subscriber has received the entire message via one or more
    RECEIVE_STREAM calls, verified that the hash matches what was provided,
    verified the authorization, and is decompressing the message
    """

    type: Literal[ReceivingState.DECOMPRESSING]
    """disciminator value"""

    task: asyncio.Task[ReceivedMessage]
    """the task that will produce the received message"""


Receiving = Union[
    ReceivingIncomplete,
    ReceivingAuthorizingMissed,
    ReceivingAuthorizing,
    ReceivingWaitingCompressor,
    ReceivingDecompressing,
]


class SendingState(Enum):
    """Discriminator value for `StateOpen.sending`"""

    SIMPLE = auto()
    """The subscriber is sending a simple message and no additional cleanup work
    is needed
    """

    MANAGEMENT_TASK = auto()
    """The subscriber is sending a management task; this may take multiple event
    loops to build the authorization part of the message
    """

    INTERNAL_MESSAGE = auto()
    """The subscriber is sending an internal message which MIGHT be in the ack queue,
    but it also might not be. When cleaning up, if the message isn't found in the ack
    queue (by identifier), it should be treated as a SENT notification and have its
    status callback called
    """


@fast_dataclass
class SendingSimple:
    """The subscriber is sending a message that was entirely deduced before being sent
    and did not involve acks
    """

    type: Literal[SendingState.SIMPLE]
    """discriminator value"""
    task: asyncio.Task[None]
    """the task that is sending the message"""


@fast_dataclass
class SendingManagementTask:
    """The subscriber is performing a management task; the management task is
    added to `expected_acks` before the task is started, meaning it can be
    acknowledged normally before this completes, without a need to delay as
    the state is not ever ambiguous
    """

    type: Literal[SendingState.MANAGEMENT_TASK]
    """discriminator value"""
    management_task: ManagementTask
    """the management task that is being sent"""
    task: asyncio.Task[None]
    """the task that is sending the message"""


@fast_dataclass
class SendingInternalMessage:
    """The subscriber is sending an internal message that may or may not be in the
    ack queue
    """

    type: Literal[SendingState.INTERNAL_MESSAGE]
    """discriminator value"""
    task: asyncio.Task[None]
    """the task that is sending the message"""
    internal_message: InternalMessage
    """the message that is being sent

    ## callback

    if the task is not done or failed, the callback could think it is in any of these states:
    - UNSENT
    - RESENDING
    - SENT

    if the task is done and successful, the callback will think it is in SENT. to ensure
    this is the case, the read task will not progress if it receives an acknowledgement
    for the message that is still in sending, as otherwise it would be possible the callback
    thinks the message is in ACKNOWLEDGED, which would be an issue when recovering (ACKNOWLEDGED
    must never change state to RESENDING)

    ## location in state

    if the task is not done or raised an error, the message may be nowhere. otherwise, it
    will be in expected_acks and sent_notifications. this can only be guarranteed because
    the read task will not progress if it receives an acknowledgement for the message that
    is still in sending
    """


Sending = Union[SendingSimple, SendingManagementTask, SendingInternalMessage]


@dataclass
class UnsentAckContinueReceive:
    """Request the broadcaster continue sending RECEIVE_STREAM"""

    type: Literal[SubscriberToBroadcasterStatefulMessageType.CONTINUE_RECEIVE]
    """discriminator value"""

    identifier: bytes
    """the message being received"""

    part_id: int
    """the last part id received"""


@dataclass
class UnsentAckConfirmReceive:
    """Confirms receipt of a message from the broadcaster"""

    type: Literal[SubscriberToBroadcasterStatefulMessageType.CONFIRM_RECEIVE]
    """discriminator value"""

    identifier: bytes
    """the message received"""


UnsentAck = Union[UnsentAckContinueReceive, UnsentAckConfirmReceive]


@dataclass
class StateConnecting:
    """the variables when in the CONNECTING state"""

    type: Literal[StateType.CONNECTING]
    """discriminator value"""

    config: WebsocketPubSubConfig
    """how the subscriber is configured"""

    client_session: aiohttp.ClientSession
    """the client session the websocket is being connected by means of"""

    websocket_task: asyncio.Task[ClientWebSocketResponse]
    """the task that is connecting to the broadcaster"""

    cancel_requested: asyncio.Event
    """if set, the state machine should move to closed as soon as possible"""

    broadcaster: PubSubBroadcasterConfig
    """the broadcaster that is being connected to"""

    retry: RetryInformation
    """information required for proceeding in the retry process"""

    tasks: TasksOnceOpen
    """the tasks that need to be performed after configuring the stream"""

    backgrounded: Set[asyncio.Task[Any]]
    """
    tasks that have been scheduled and if they error it's not recoverable, but
    the result isnt otherwise important. the most prominent example is informing
    callbacks on internal messages

    these shouldnt be canceled unless the state machine is moving to CLOSED; i.e.,
    retries should not cause these tasks to be canceled
    """


@dataclass
class StateConfiguring:
    """the variables when in the CONFIGURING state"""

    type: Literal[StateType.CONFIGURING]
    """discriminator value"""

    client_session: ClientSession
    """the client session the websocket is part of"""

    config: WebsocketPubSubConfig
    """how the subscriber is configured"""

    cancel_requested: asyncio.Event
    """if set, the state machine should move to closed as soon as possible"""

    broadcaster: PubSubBroadcasterConfig
    """the broadcaster that the websocket goes to"""

    websocket: ClientWebSocketResponse
    """the live websocket to the broadcaster"""

    retry: RetryInformation
    """information required for proceeding in the retry process"""

    tasks: TasksOnceOpen
    """the tasks that need to be performed after configuring the stream"""

    subscriber_nonce: bytes
    """the 32 bytes that the subscriber is contributing to the connection nonce"""

    send_task: Optional[asyncio.Task[None]]
    """if still trying to send the configure message, the task for sending it"""

    read_task: asyncio.Task[WSMessage]
    """the task for reading the next message from the websocket"""

    backgrounded: Set[asyncio.Task[Any]]
    """
    tasks that have been scheduled and if they error it's not recoverable, but
    the result isnt otherwise important. the most prominent example is informing
    callbacks on internal messages

    these shouldnt be canceled unless the state machine is moving to CLOSED; i.e.,
    retries should not cause these tasks to be canceled
    """


@dataclass
class StateOpen:
    """the variables when in the OPEN state"""

    type: Literal[StateType.OPEN]
    """discriminator value"""

    client_session: ClientSession
    """the client session the websocket is part of"""

    config: WebsocketPubSubConfig
    """how the subscriber is configured"""

    cancel_requested: asyncio.Event
    """if set, the state machine should move to closed as soon as possible"""

    broadcaster: PubSubBroadcasterConfig
    """the broadcaster that the websocket is connected to"""

    broadcaster_counter: int
    """A counter that is incremented whenever the broadcaster could use a url to
    generate an authorization code
    """

    subscriber_counter: int
    """A counter that is decremented whenever the subscriber could use a url to
    generate an authorization code
    """

    nonce_b64: str
    """the agreed unique identifier for this connection, that combines the
    subscribers contribution and the broadcasters contribution, both of which
    are asked to produce 32 random bytes

    base64 encoded since thats how it is used
    """

    websocket: ClientWebSocketResponse
    """the live websocket to the broadcaster"""

    retry: OpenRetryInformation
    """information required for proceeding in the retry process"""

    compressors: CompressorStore
    """the available compressors for decompressing incoming messages or compressing
    outgoing messages
    """

    unsent_notifications: DrainableAsyncioQueue[InternalMessage]
    """the unsent messages that should be sent to the broadcaster via NOTIFY / NOTIFY STREAM
    in this order. Since these notifications are in UNSENT with nothing queued it's not
    necessary to sweep the internal message state callbacks for these messages
    """

    resending_notifications: List[InternalMessage]
    """the resending messages that should be sent to the broadcaster via NOTIFY / NOTIFY STREAM
    in this order. the internal message state callbacks need to be swept consistently

    NOTE: the subscriber never retries a message within the same websocket connection,
    so it cannot receive acknowledgements for these messages. if the broadcaster does
    not acknowledge a notification in time the connection is closed. thus, the distinction
    between unsent and resending here is only to distinguish which need to be swept and
    which don't

    NOTE: resending notifications are always handled before sent notifications, so this
    empties out at the beginning of the connection and never refills, meaning the linear
    pass through this list is generally a no-op
    """

    sent_notifications: BoundedDeque[InternalMessage]
    """the messages that have been sent to the broadcaster but not acknowledged,
    in the order they are expected to be acknowledged. the last message in this
    list may only have been partially sent. these need to be swept regularly

    NOTE: this list is just a subset of `expected_acks` in the following sense:
        - `len(sent_notifications) <= len(expected_acks)`
        - for every message in `sent_notifications`, there is a corresponding
          acknowledgement in `expected_acks`
        - the order of the messages in `sent_notifications` is the same as the order
          of the corresponding acknowledgements in `expected_acks`

    NOTE: the broadcaster is given until `expected_acks` reaches its max size or
    the websocket ping determines the connection is dead to acknowledge a message;
    there is otherwise no timeout. thus in theory a malicious broadcaster could
    cause a DoS for the client, though this scenario seems unlikely
    """

    exact_subscriptions: Set[bytes]
    """the topics the subscriber is subscribed to BEFORE all management tasks/acks; 
    this is used for restoring the state if the connection is lost
    """

    glob_subscriptions: Set[str]
    """the glob patterns the subscriber is susbcribed to BEFORE all management tasks/acks
    are completed; this is used for restoring the state if the connection is lost
    """

    management_tasks: DrainableAsyncioQueue[ManagementTask]
    """the management tasks that need to be performed in the order they need to be
    performed; these have not had their ack added to expected_acks and have not
    been applied to `exact_subscriptions`/`glob_subscriptions` yet
    """

    expected_acks: BoundedDeque[Acknowledgement]
    """the acknowledgements the subscriber expects to receive in the order they
    are expected to be received; receiving an acknowledgement out of order 
    is an error.
    """

    receiving: Optional[Receiving]
    """if the subscriber has received some part of a notification
    but knows theres more to come or hasn't finished processing it, the state of
    that receiving notification, otherwise None. the broadcaster MUST always
    finish the last notification before sending a new one, and the subscriber
    does not try to read additional messages while processing the previous
    receive stream, so only one notification can be in this state at a time.
    """

    unsent_acks: BoundedDeque[UnsentAck]
    """Outgoing acknowledgements that need to be sent"""

    received: DrainableAsyncioQueue[ReceivedMessage]
    """the messages that have been received from the broadcaster but not yet
    consumed; it's expected that this is consumed from outside the state machine
    """

    sending: Optional[Sending]
    """the task which has exclusive access to `send_bytes`, if any, otherwise
    None
    """

    read_task: asyncio.Task[WSMessage]
    """the task responsible for reading the next message from the websocket"""

    backgrounded: Set[asyncio.Task[Any]]
    """
    tasks that have been scheduled and if they error it's not recoverable, but
    the result isnt otherwise important. the most prominent example is informing
    callbacks on internal messages

    these shouldnt be canceled unless the state machine is moving to CLOSED; i.e.,
    retries should not cause these tasks to be canceled
    """


@dataclass
class StateWaitingRetry:
    """the variables in the WAITING_RETRY state"""

    type: Literal[StateType.WAITING_RETRY]
    """discriminator value"""

    config: WebsocketPubSubConfig
    """how the subscriber is configured"""

    cancel_requested: asyncio.Event
    """if set, the state machine should move to closed as soon as possible"""

    retry: RetryInformation
    """information required for proceeding in the retry process"""

    tasks: TasksOnceOpen
    """the tasks that need to be performed after configuring the stream"""

    retry_at: float
    """the time in fractional seconds from the epoch (as if by `time.time()`)
    before proceeding to the next attempt
    """

    backgrounded: Set[asyncio.Task[Any]]
    """
    tasks that have been scheduled and if they error it's not recoverable, but
    the result isnt otherwise important. the most prominent example is informing
    callbacks on internal messages

    these shouldnt be canceled unless the state machine is moving to CLOSED; i.e.,
    retries should not cause these tasks to be canceled
    """


@dataclass
class StateClosing:
    """the variables in the CLOSING state"""

    type: Literal[StateType.CLOSING]
    """discriminator value"""

    config: WebsocketPubSubConfig
    """how the subscriber is configured"""

    cancel_requested: asyncio.Event
    """if set, the state machine should move to closed as soon as possible"""

    broadcaster: PubSubBroadcasterConfig
    """the broadcaster that the websocket was connected to"""

    client_session: ClientSession
    """the client session the websocket is part of"""

    websocket: ClientWebSocketResponse
    """the potentially live websocket to the broadcaster"""

    retry: ClosingRetryInformation
    """determines if and how the subscriber will retry connecting to
    a broadcaster once the websocket is done closing
    """

    backgrounded: Set[asyncio.Task[Any]]
    """
    tasks that have been scheduled and if they error it's not recoverable, but
    the result isnt otherwise important. the most prominent example is informing
    callbacks on internal messages

    these shouldnt be canceled unless the state machine is moving to CLOSED; i.e.,
    retries should not cause these tasks to be canceled
    """


@fast_dataclass
class StateClosed:
    """the variables in the CLOSED state"""

    type: Literal[StateType.CLOSED]
    """discriminator value"""


State = Union[
    StateConnecting,
    StateConfiguring,
    StateOpen,
    StateWaitingRetry,
    StateClosing,
    StateClosed,
]



import asyncio
from typing import TYPE_CHECKING, Dict, List, Optional, Protocol, Tuple, Type

from lonelypsp.auth.config import AuthConfig, AuthResult
from lonelypsp.auth.set_subscriptions_info import SetSubscriptionsInfo
from lonelypsp.stateful.messages.configure import S2B_Configure
from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigure
from lonelypsp.stateful.messages.continue_notify import B2S_ContinueNotify
from lonelypsp.stateful.messages.continue_receive import S2B_ContinueReceive
from lonelypsp.stateful.messages.disable_zstd_custom import B2S_DisableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_custom import B2S_EnableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_preset import B2S_EnableZstdPreset
from lonelypsp.stateless.make_strong_etag import StrongEtag

from lonelypsc.config.config import PubSubBroadcasterConfig

try:
    import zstandard
except ImportError:
    ...


class WebsocketPubSubConnectConfig(Protocol):
    """Determines how we connect to broadcasters"""

    @property
    def broadcasters(self) -> List[PubSubBroadcasterConfig]:
        """The broadcasters that we can connect to for making subscription requests
        and requesting a message be broadcast
        """

    @property
    def outgoing_initial_connect_retries(self) -> int:
        """The number of times we will retry the initial connection to a broadcaster.
        We will try every broadcaster at least once before retrying the same one
        """

    @property
    def outgoing_min_reconnect_interval(self) -> float:
        """Given that we successfully complete all the initial handshakes with a broadcaster
        and are satisfied the connection is up, how long do we need to then stay alive before
        we consider the connection stable.

        Most easily understood by example:

        Suppose we have broadcasters 1, 2, and 3. We connect to 1, it works, but 12 hours later
        we encounter an error. We reconnect to 2, it stays alive 12 hours, then errors. We connect
        to 3, it stays alive 12 hours, then errors. It seems reasonable that things are working well
        enough that we are ok to just go back to 1, since even if it does die again in 12h it's a
        perfectly manageable amount of overhead.

        Alternatively, if we connected to 1, completed the initial handshake, and died 3s later, then
        connected to 2, same thing, 3, same thing, probably we should stop trying as we're spending a
        lot of time managing connections compared to actually using them.

        Thus, following the above argument, the min reconnect interval should be between 3s and 12h.
        Generally, you should choose a value low enough that if it was reconnecting that often you
        would want to be pinged about it, since this is going to raise errors which will presumably
        trigger your alerting mechanisms.
        """


class WebsocketPubSubConnectConfigFromParts:
    """Convenience class to construct an object fulfilling the WebsocketPubSubConnectConfig
    protocol
    """

    def __init__(
        self,
        broadcasters: List[PubSubBroadcasterConfig],
        outgoing_initial_connect_retries: int,
        outgoing_min_reconnect_interval: float,
    ):
        self.broadcasters = broadcasters
        self.outgoing_initial_connect_retries = outgoing_initial_connect_retries
        self.outgoing_min_reconnect_interval = outgoing_min_reconnect_interval


if TYPE_CHECKING:
    _: Type[WebsocketPubSubConnectConfig] = WebsocketPubSubConnectConfigFromParts


class WebsocketGenericConfig(Protocol):
    @property
    def max_websocket_message_size(self) -> Optional[int]:
        """The maximum size in bytes for outgoing websocket messages. In theory,
        websocket messages are already broken up with websocket frames, which are
        then broken up with TCP packets, so it's redundant to have this limit.
        In practice, intermediaries will either drop large messages or behave
        poorly when they receive them. Thus, for maximum compatibility, set this
        to 16mb or less.
        """

    @property
    def websocket_open_timeout(self) -> Optional[float]:
        """The maximum amount of time to wait for the websocket connection to be
        established before trying the next broadcaster
        """

    @property
    def websocket_close_timeout(self) -> Optional[float]:
        """The maximum amount of time to wait after trying to close the websocket
        connection for the acknowledgement from the broadcaster
        """

    @property
    def websocket_heartbeat_interval(self) -> float:
        """If the websocket is idle for this duration in seconds, the subscriber
        sends a PING frame and waits up to half this duration to receive
        a PONG frame or the subscriber assumes the connection is lost and
        moves to the next broadcaster. A lower value causes more overhead but
        more quickly detects connection issues.

        On the broadcaster side this is configured by the ASGI server, e.g.,
        uvicorn has `--ws-ping-interval <float> --ws-ping-timeout <float>`.
        The lower of the two ping intervals is the effective ping overhead
        of a healthy connection, but both ends must still have a heartbeat
        interval (not necessarily the same) to detect a failure, since they
        may not receive the FIN from the other end if the connection is lost

        It is not recommended to have the same ping interval on both sides as
        it leads to needlessly doubled PING/PONG (both sides will send a ping
        before receiving the corresponding pong), so a good choice is e.g.
        subscriber has X, broadcasters >=2X, since subscribers are more sensitive
        to lost connections depending on the usecase and may want to tune it lower
        if it's very important they detect missed messages quickly, but can reduce
        overhead if it's not as important, whereas if the broadcaster side is low
        all connections have high overhead

        NOTE: default uvicorn settings is ping interval 20, ping timeout 20, so
        a good default choice is a heartbeat interval 10 for subscribers, with
        a value no higher than 19 to avoid doubling and no lower than the 2x the
        99.9% roundtrip time to reduce superfluous reconnects
        """

    @property
    def websocket_minimal_headers(self) -> bool:
        """True if all messages from the subscriber to the broadcaster should use
        minimal headers, which are faster to parse and more compact but require
        that the subscriber and broadcaster precisely agree on the headers for
        each message. False if all messages from the subscriber to the
        broadcaster use expanded headers, which are more flexible and easier to
        debug but slower to parse and more verbose.

        If you are trying to understand the lonelypss protocol via something
        like wireshark, setting this to False will make messages somewhat easier
        to understand.

        Note that broadcasters and subscribers do not need to agree on this
        setting. It is ok if the broadcaster is sending expanded headers and the
        subscriber is sending minimal headers, or vice versa, as this only
        configures the outgoing messages but they both always accept either
        version for incoming messages.

        Generally, this should be True except when in the process of updating
        the lonelypss/lonelypsc libraries, in which case it should be changed to
        false on the broadcaster and subscribers, then they should be updated
        one at a time, then set to true.
        """

    @property
    def max_sent_notifications(self) -> Optional[int]:
        """The maximum number of unacknowledged notifications before the subscriber
        disconnects because the broadcaster cannot keep up, or None for no limit.

        This is not useful for backpressure; instead, use the fact that notify()
        returns a coroutine that is not complete until the broadcaster has
        acknowledged the notification or the subscriber has dropped it, so e.g.
        a semaphore (or any other concurrency primitive) can be used to limit the
        number of unacknowledged notifications

        This is useful as a better error message if the above mechanism is not
        working as intended, so e.g. if the caller knows that it intends to have
        at most 3 unacknowledged notifications at any time, it can set this to 3
        so that an error is raised if there are 4 unacknowledged notifications
        """

    @property
    def max_unsent_notifications(self) -> Optional[int]:
        """the maximum number of unsent notifications queued up before the subscriber
        disconnects because either the broadcaster or the subscriber cannot keep
        up, or None for no limit.

        This is very similar to `max_sent_notifications` in practice, and should usually
        be set to the same value. This could differ in theory in that it will trigger
        first if trying to send a lot of notifications within the same event loop, or
        the network buffer is actually full (i.e., `send_bytes` is taking a long time)
        """

    @property
    def max_expected_acks(self) -> Optional[int]:
        """The maximum number of unacknowledged management tasks OR notifications;
        this should be at least `max_sent_notifications` plus the number of subscriptions
        (topic or glob) that may be made

        This is not useful for backpressure, and there is no way to implement
        backpressure for subscribe requests as they are sent in bulk at the
        beginning of a retry (since it would be highly unusual to have so many
        that backpressure is required). Instead, if there are a lot of topics
        the subscriber is interested in, use glob patterns to reduce the amount
        of work by the broadcaster and noise when reconnecting

        This is useful as a sanity check that the subscriber is not sending an
        excessive number of subscribe requests at the start of the connection
        (e.g., thousands), which likely means there is a bug or poor pattern in
        the subscriber code

        Note that if you DO have a good reason for subscribing to many topics
        instead of combining them with a pattern, there is no sudden cutoff when
        setting this to a large value (or None). The subscriptions all need to
        be maintained in memory on the subscribers side; so e.g. if there are
        100,000 topics at 64 bytes each it would take on the order of 6mb of
        memory for the topics themselves, plus 2-5x that when reconnecting. It
        may also be helpful, if doing this, to look over the DB config on the
        broadcaster and see if there are possible optimizations on how the
        subscriptions are stored/retrieved for this usecase

        Performance wise, having lots of exact topic subscriptions generally
        doesn't incur excessive overhead (as it's a dict lookup on the subscriber
        side and a btree lookup on the broadcaster side), but having lots of
        glob subscriptions generally does incur linear overhead on both sides.
        If you have a lot of glob subscriptions (>100), but not an excessive
        number of topics, consider at least changing the broadcasters
        implementation to cache topic -> current globs that match to speed up
        that side (this is generally not a good optimization if there are not
        a significant number of glob subscriptions)
        """

    @property
    def max_received(self) -> Optional[int]:
        """The maximum number of messages that have been received by the
        subscriber but not yet sent to message receivers, or None for no
        limit. At this limit the websocket will be disconnected.

        This is not useful for backpressure as the websocket gets disconnected
        when this limit is reached; instead, it is useful as a sanity check
        to ensure the subscriber can keep up with the incoming messages.
        """

    @property
    def max_unsent_acks(self) -> Optional[int]:
        """The maximum number of unsent acknowledgements from the subscriber
        to the broadcaster before disconnecting because connection cannot keep
        up

        This is an unlikely failure point since the broadcaster is expected to
        be waiting for acknowledgements before sending an excessive number of
        messages, and thus this is intended just as a sanity check
        """


class WebsocketGenericConfigFromParts:
    """Convenience class to construct an object fulfilling the WebsocketGenericConfig
    protocol
    """

    def __init__(
        self,
        max_websocket_message_size: Optional[int],
        websocket_open_timeout: Optional[float],
        websocket_close_timeout: Optional[float],
        websocket_heartbeat_interval: float,
        websocket_minimal_headers: bool,
        max_sent_notifications: Optional[int],
        max_unsent_notifications: Optional[int],
        max_expected_acks: Optional[int],
        max_received: Optional[int],
        max_unsent_acks: Optional[int],
    ):
        self.max_websocket_message_size = max_websocket_message_size
        self.websocket_open_timeout = websocket_open_timeout
        self.websocket_close_timeout = websocket_close_timeout
        self.websocket_heartbeat_interval = websocket_heartbeat_interval
        self.websocket_minimal_headers = websocket_minimal_headers
        self.max_sent_notifications = max_sent_notifications
        self.max_unsent_notifications = max_unsent_notifications
        self.max_expected_acks = max_expected_acks
        self.max_received = max_received
        self.max_unsent_acks = max_unsent_acks


if TYPE_CHECKING:
    __: Type[WebsocketGenericConfig] = WebsocketGenericConfigFromParts


class WebsocketCompressionConfig(Protocol):
    @property
    def allow_compression(self) -> bool:
        """True to enable zstandard compression within the websocket connection, False
        to disable it
        """

    async def get_compression_dictionary_by_id(
        self, dictionary_id: int, /, *, level: int
    ) -> "Optional[zstandard.ZstdCompressionDict]":
        """If a precomputed zstandard compression dictionary is available with the
        given id, the corresponding dictionary should be returned. Presets must
        already be available on the broadcaster in order to be used. They
        provide meaningful compression on small messages (where a compression dict
        is too large to send alongside the message) when using short-lived websockets
        (where there isn't enough time to train a dictionary)

        The provided compression level is the hint returned from the broadcaster,
        to avoid having to duplicate that configuration here. The returned dict
        should have its data precomputed as if by `precompute_compress`
        """

    @property
    def initial_compression_dict_id(self) -> Optional[int]:
        """Either None to indicate no initial preset compression dictionary should be
        used, or the id of the preset compression dictionary appropriate for the
        messages that are expected to be sent over the connection. The subscriber will
        send this recommendation in the CONFIGURE packet, but will not use it unless
        the broadcaster agrees with the ENABLE_ZSTD_PRESET packet, which will also
        configure the other metadata (min/max size of messages to use this preset with,
        compression level)
        """

    @property
    def allow_training_compression(self) -> bool:
        """True to allow the broadcaster to train a custom zstandard dict on the small
        payloads that we receive or send over the websocket connection, then
        send us that custom dictionary so we can reuse it for better
        compression.

        The broadcaster may be configured differently, but typically it will train
        on messages between 32 and 16384 bytes, which is large enough that
        compression with a pre-trained dictionary may help, but small enough
        that the the overhead of providing a dictionary alongside each message
        would overwhelm the benefits of compression.

        Generally the subscriber should enable this if it expects to send/receive enough
        relevant messages to reach the training thresholds (usually 100kb to 1mb
        of relevant messages), plus enough to make the training overhead worth
        it (typically another 10mb or so). The subscriber should disable this if it won't
        send relevant messages or it expects to disconnect before sending/receiving
        enough data for the training to complete or the savings to compensate for
        the work spent building the dictionary.

        The subscriber should also disable this if the message payloads will not be
        meaningfully compressible, e.g., significant parts are random or encrypted data.
        Generally, for encryption, TLS should be used so that compression can still
        occur on the unencrypted payload (i.e., raw -> compressed -> encrypted).

        The automatically trained compression will generally convert a simple
        protocol design, such as json with long key names and extra wrappers for
        extensibility, into the same order of magnitude network size as a more
        compact protocol
        """

    @property
    def decompression_max_window_size(self) -> int:
        """
        Sets an upper limit on the window size for decompression operations
        in kibibytes. This setting can be used to prevent large memory
        allocations for inputs using large compression windows.

        Use 0 for no limit.

        A reasonable value is 0 for no limit. Alternatively, it should be 8mb if
        trying to match the zstandard minimum decoder requirements. The
        remaining alternative would be as high as the subscriber can bear

        WARN:
            This should not be considered a security measure. Authorization
            is already passed prior to decompression, and if that is not enough
            to eliminate adversarial payloads, then disable compression.
        """


class WebsocketCompressionConfigFromParts:
    """Convenience class to construct an object fulfilling the WebsocketCompressionConfig
    protocol
    """

    def __init__(
        self,
        allow_compression: bool,
        compression_dictionary_by_id: "Dict[int, List[Tuple[int, zstandard.ZstdCompressionDict]]]",
        initial_compression_dict_id: Optional[int],
        allow_training_compression: bool,
        decompression_max_window_size: int,
    ):
        self.allow_compression = allow_compression
        self.compression_dictionary_by_id = compression_dictionary_by_id
        """
        Maps from dictionary id to a sorted list of (level, precomputed dictionary). You should
        initialize this with a guess for what level the broadcaster will suggest for compression,
        typically 10, and this will automatically fill in remaining levels as needed.
        """
        self.initial_compression_dict_id = initial_compression_dict_id
        self.allow_training_compression = allow_training_compression
        self.decompression_max_window_size = decompression_max_window_size

    async def get_compression_dictionary_by_id(
        self, dictionary_id: int, /, *, level: int
    ) -> "Optional[zstandard.ZstdCompressionDict]":
        sorted_choices = self.compression_dictionary_by_id.get(dictionary_id, None)
        if not sorted_choices:
            return None

        for insert_idx, (choice_level, dictionary) in enumerate(sorted_choices):
            if choice_level == level:
                return dictionary
            elif choice_level > level:
                break

        data = sorted_choices[0][1].as_bytes()
        zdict = zstandard.ZstdCompressionDict(data)
        await asyncio.to_thread(zdict.precompute_compress, level)
        sorted_choices.insert(insert_idx, (level, zdict))
        return zdict


if TYPE_CHECKING:
    ___: Type[WebsocketCompressionConfig] = WebsocketCompressionConfigFromParts


class WebsocketPubSubConfig(
    WebsocketPubSubConnectConfig,
    WebsocketGenericConfig,
    WebsocketCompressionConfig,
    AuthConfig,
    Protocol,
): ...


class WebsocketPubSubConfigFromParts:
    """Convenience class to construct an object fulfilling the WebsocketPubSubConfig
    protocol from objects which fulfill the various parts
    """

    def __init__(
        self,
        connect: WebsocketPubSubConnectConfig,
        generic: WebsocketGenericConfig,
        compression: WebsocketCompressionConfig,
        auth: AuthConfig,
    ):
        self.connect = connect
        self.generic = generic
        self.compression = compression
        self.auth = auth

    @property
    def broadcasters(self) -> List[PubSubBroadcasterConfig]:
        return self.connect.broadcasters

    @property
    def outgoing_initial_connect_retries(self) -> int:
        return self.connect.outgoing_initial_connect_retries

    @property
    def outgoing_min_reconnect_interval(self) -> float:
        return self.connect.outgoing_min_reconnect_interval

    @property
    def max_websocket_message_size(self) -> Optional[int]:
        return self.generic.max_websocket_message_size

    @property
    def websocket_open_timeout(self) -> Optional[float]:
        return self.generic.websocket_open_timeout

    @property
    def websocket_close_timeout(self) -> Optional[float]:
        return self.generic.websocket_close_timeout

    @property
    def websocket_heartbeat_interval(self) -> float:
        return self.generic.websocket_heartbeat_interval

    @property
    def websocket_minimal_headers(self) -> bool:
        return self.generic.websocket_minimal_headers

    @property
    def max_sent_notifications(self) -> Optional[int]:
        return self.generic.max_sent_notifications

    @property
    def max_unsent_notifications(self) -> Optional[int]:
        return self.generic.max_unsent_notifications

    @property
    def max_expected_acks(self) -> Optional[int]:
        return self.generic.max_expected_acks

    @property
    def max_received(self) -> Optional[int]:
        return self.generic.max_received

    @property
    def max_unsent_acks(self) -> Optional[int]:
        return self.generic.max_unsent_acks

    @property
    def allow_compression(self) -> bool:
        return self.compression.allow_compression

    async def get_compression_dictionary_by_id(
        self, dictionary_id: int, /, *, level: int
    ) -> "Optional[zstandard.ZstdCompressionDict]":
        return await self.compression.get_compression_dictionary_by_id(
            dictionary_id, level=level
        )

    @property
    def initial_compression_dict_id(self) -> Optional[int]:
        return self.compression.initial_compression_dict_id

    @property
    def allow_training_compression(self) -> bool:
        return self.compression.allow_training_compression

    @property
    def decompression_max_window_size(self) -> int:
        return self.compression.decompression_max_window_size

    async def setup_to_broadcaster_auth(self) -> None:
        await self.auth.setup_to_broadcaster_auth()

    async def teardown_to_broadcaster_auth(self) -> None:
        await self.auth.teardown_to_broadcaster_auth()

    async def setup_to_subscriber_auth(self) -> None:
        await self.auth.setup_to_subscriber_auth()

    async def teardown_to_subscriber_auth(self) -> None:
        await self.auth.teardown_to_subscriber_auth()

    async def authorize_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_subscribe_exact(
            tracing=tracing, url=url, recovery=recovery, exact=exact, now=now
        )

    async def is_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_subscribe_exact_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            exact=exact,
            now=now,
            authorization=authorization,
        )

    async def authorize_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_subscribe_glob(
            tracing=tracing, url=url, recovery=recovery, glob=glob, now=now
        )

    async def is_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_subscribe_glob_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            glob=glob,
            now=now,
            authorization=authorization,
        )

    async def authorize_notify(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_notify(
            tracing=tracing,
            topic=topic,
            identifier=identifier,
            message_sha512=message_sha512,
            now=now,
        )

    async def is_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_notify_allowed(
            tracing=tracing,
            topic=topic,
            identifier=identifier,
            message_sha512=message_sha512,
            now=now,
            authorization=authorization,
        )

    async def authorize_stateful_configure(
        self,
        /,
        *,
        tracing: bytes,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_configure(
            tracing=tracing,
            subscriber_nonce=subscriber_nonce,
            enable_zstd=enable_zstd,
            enable_training=enable_training,
            initial_dict=initial_dict,
        )

    async def is_stateful_configure_allowed(
        self, /, *, message: S2B_Configure, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_configure_allowed(message=message, now=now)

    async def authorize_check_subscriptions(
        self, /, *, tracing: bytes, url: str, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_check_subscriptions(
            tracing=tracing, url=url, now=now
        )

    async def is_check_subscriptions_allowed(
        self, /, *, tracing: bytes, url: str, now: float, authorization: Optional[str]
    ) -> AuthResult:
        return await self.auth.is_check_subscriptions_allowed(
            tracing=tracing, url=url, now=now, authorization=authorization
        )

    async def authorize_set_subscriptions(
        self, /, *, tracing: bytes, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_set_subscriptions(
            tracing=tracing, url=url, strong_etag=strong_etag, now=now
        )

    async def is_set_subscriptions_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        strong_etag: StrongEtag,
        subscriptions: SetSubscriptionsInfo,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_set_subscriptions_allowed(
            tracing=tracing,
            url=url,
            strong_etag=strong_etag,
            subscriptions=subscriptions,
            now=now,
            authorization=authorization,
        )

    async def authorize_stateful_continue_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        part_id: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_continue_receive(
            tracing=tracing, identifier=identifier, part_id=part_id, url=url, now=now
        )

    async def is_stateful_continue_receive_allowed(
        self, /, *, url: str, message: S2B_ContinueReceive, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_continue_receive_allowed(
            url=url, message=message, now=now
        )

    async def authorize_confirm_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_confirm_receive(
            tracing=tracing,
            identifier=identifier,
            num_subscribers=num_subscribers,
            url=url,
            now=now,
        )

    async def is_confirm_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_confirm_receive_allowed(
            tracing=tracing,
            identifier=identifier,
            num_subscribers=num_subscribers,
            url=url,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_missed(
        self, /, *, tracing: bytes, topic: bytes, url: str, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_confirm_missed(
            tracing=tracing, topic=topic, url=url, now=now
        )

    async def is_confirm_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_confirm_missed_allowed(
            tracing=tracing, topic=topic, url=url, now=now, authorization=authorization
        )

    async def authorize_receive(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_receive(
            tracing=tracing,
            url=url,
            topic=topic,
            message_sha512=message_sha512,
            identifier=identifier,
            now=now,
        )

    async def is_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_receive_allowed(
            tracing=tracing,
            url=url,
            topic=topic,
            message_sha512=message_sha512,
            identifier=identifier,
            now=now,
            authorization=authorization,
        )

    async def authorize_missed(
        self, /, *, tracing: bytes, recovery: str, topic: bytes, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_missed(
            tracing=tracing, recovery=recovery, topic=topic, now=now
        )

    async def is_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_missed_allowed(
            tracing=tracing,
            recovery=recovery,
            topic=topic,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_confirm_subscribe_exact(
            tracing=tracing, url=url, recovery=recovery, exact=exact, now=now
        )

    async def is_confirm_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_confirm_subscribe_exact_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            exact=exact,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_confirm_subscribe_glob(
            tracing=tracing, url=url, recovery=recovery, glob=glob, now=now
        )

    async def is_confirm_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_confirm_subscribe_glob_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            glob=glob,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_notify(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_confirm_notify(
            tracing=tracing,
            identifier=identifier,
            subscribers=subscribers,
            topic=topic,
            message_sha512=message_sha512,
            now=now,
        )

    async def is_confirm_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        authorization: Optional[str],
        now: float,
    ) -> AuthResult:
        return await self.auth.is_confirm_notify_allowed(
            tracing=tracing,
            identifier=identifier,
            subscribers=subscribers,
            topic=topic,
            message_sha512=message_sha512,
            authorization=authorization,
            now=now,
        )

    async def authorize_check_subscriptions_response(
        self,
        /,
        *,
        tracing: bytes,
        strong_etag: StrongEtag,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_check_subscriptions_response(
            tracing=tracing, strong_etag=strong_etag, now=now
        )

    async def is_check_subscription_response_allowed(
        self,
        /,
        *,
        tracing: bytes,
        strong_etag: StrongEtag,
        authorization: Optional[str],
        now: float,
    ) -> AuthResult:
        return await self.auth.is_check_subscription_response_allowed(
            tracing=tracing,
            strong_etag=strong_etag,
            authorization=authorization,
            now=now,
        )

    async def authorize_set_subscriptions_response(
        self, /, *, tracing: bytes, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_set_subscriptions_response(
            tracing=tracing, strong_etag=strong_etag, now=now
        )

    async def is_set_subscription_response_allowed(
        self,
        /,
        *,
        tracing: bytes,
        strong_etag: StrongEtag,
        authorization: Optional[str],
        now: float,
    ) -> AuthResult:
        return await self.auth.is_set_subscription_response_allowed(
            tracing=tracing,
            strong_etag=strong_etag,
            authorization=authorization,
            now=now,
        )

    async def authorize_stateful_confirm_configure(
        self, /, *, broadcaster_nonce: bytes, tracing: bytes, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_confirm_configure(
            broadcaster_nonce=broadcaster_nonce, tracing=tracing, now=now
        )

    async def is_stateful_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_confirm_configure_allowed(
            message=message, now=now
        )

    async def authorize_stateful_enable_zstd_preset(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_enable_zstd_preset(
            tracing=tracing,
            url=url,
            compressor_identifier=compressor_identifier,
            compression_level=compression_level,
            min_size=min_size,
            max_size=max_size,
            now=now,
        )

    async def is_stateful_enable_zstd_preset_allowed(
        self, /, *, url: str, message: B2S_EnableZstdPreset, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_enable_zstd_preset_allowed(
            url=url, message=message, now=now
        )

    async def authorize_stateful_enable_zstd_custom(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_enable_zstd_custom(
            tracing=tracing,
            url=url,
            compressor_identifier=compressor_identifier,
            compression_level=compression_level,
            min_size=min_size,
            max_size=max_size,
            sha512=sha512,
            now=now,
        )

    async def is_stateful_enable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_EnableZstdCustom, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_enable_zstd_custom_allowed(
            url=url, message=message, now=now
        )

    async def authorize_stateful_disable_zstd_custom(
        self, /, *, tracing: bytes, compressor_identifier: int, url: str, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_disable_zstd_custom(
            tracing=tracing,
            compressor_identifier=compressor_identifier,
            url=url,
            now=now,
        )

    async def is_stateful_disable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_DisableZstdCustom, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_disable_zstd_custom_allowed(
            url=url, message=message, now=now
        )

    async def authorize_stateful_continue_notify(
        self, /, *, tracing: bytes, identifier: bytes, part_id: int, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_continue_notify(
            tracing=tracing, identifier=identifier, part_id=part_id, now=now
        )

    async def is_stateful_continue_notify_allowed(
        self, /, *, message: B2S_ContinueNotify, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_continue_notify_allowed(
            message=message, now=now
        )


if TYPE_CHECKING:
    ____: Type[WebsocketPubSubConfig] = WebsocketPubSubConfigFromParts


def make_websocket_pub_sub_config(
    broadcasters: List[PubSubBroadcasterConfig],
    outgoing_initial_connect_retries: int,
    outgoing_min_reconnect_interval: float,
    max_websocket_message_size: Optional[int],
    websocket_open_timeout: Optional[float],
    websocket_close_timeout: Optional[float],
    websocket_heartbeat_interval: float,
    websocket_minimal_headers: bool,
    max_sent_notifications: Optional[int],
    max_unsent_notifications: Optional[int],
    max_expected_acks: Optional[int],
    max_received: Optional[int],
    max_unsent_acks: Optional[int],
    allow_compression: bool,
    compression_dictionary_by_id: "Dict[int, Tuple[zstandard.ZstdCompressionDict, int]]",
    initial_compression_dict_id: Optional[int],
    allow_training_compression: bool,
    decompression_max_window_size: int,
    auth: AuthConfig,
) -> WebsocketPubSubConfig:
    """Convenience function to make a WebsocketPubSubConfig object without excessive nesting
    if you are specifying everything that doesn't need to be synced with the broadcaster
    within code.

    The compression dictionary object is inputted in the same form as the broadcaster for
    convenience, and will be converted to the appropriate form for the subscriber
    """
    return WebsocketPubSubConfigFromParts(
        connect=WebsocketPubSubConnectConfigFromParts(
            broadcasters=broadcasters,
            outgoing_initial_connect_retries=outgoing_initial_connect_retries,
            outgoing_min_reconnect_interval=outgoing_min_reconnect_interval,
        ),
        generic=WebsocketGenericConfigFromParts(
            max_websocket_message_size=max_websocket_message_size,
            websocket_open_timeout=websocket_open_timeout,
            websocket_close_timeout=websocket_close_timeout,
            websocket_heartbeat_interval=websocket_heartbeat_interval,
            websocket_minimal_headers=websocket_minimal_headers,
            max_sent_notifications=max_sent_notifications,
            max_unsent_notifications=max_unsent_notifications,
            max_expected_acks=max_expected_acks,
            max_received=max_received,
            max_unsent_acks=max_unsent_acks,
        ),
        compression=WebsocketCompressionConfigFromParts(
            allow_compression=allow_compression,
            compression_dictionary_by_id=dict(
                (dict_id, [(level, zdict)])
                for (dict_id, (zdict, level)) in compression_dictionary_by_id.items()
            ),
            allow_training_compression=allow_training_compression,
            initial_compression_dict_id=initial_compression_dict_id,
            decompression_max_window_size=decompression_max_window_size,
        ),
        auth=auth,
    )



import io
from typing import cast

from lonelypsp.stateful.parser import B2S_AnyMessageParser
from lonelypsp.stateful.parser_helpers import parse_b2s_message_prefix

from lonelypsc.client import PubSubError
from lonelypsc.types.websocket_message import WSMessageBytes
from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.handlers.open.messages.generic import check_read_message
from lonelypsc.ws.handlers.util.read_from_websocket import make_websocket_read_task
from lonelypsc.ws.state import ReceivingState, StateOpen


def check_read_task(state: StateOpen) -> CheckResult:
    """Checks if the read task is done, if so, verifies it was successful and
    the payload makes sense, otherwise raises an exception

    Skips checking the read task if still processing the previous message
    """

    if (
        state.receiving is not None
        and state.receiving.type != ReceivingState.INCOMPLETE
    ):
        return CheckResult.CONTINUE

    if not state.read_task.done():
        return CheckResult.CONTINUE

    raw_result = state.read_task.result()
    if raw_result["type"] == "websocket.disconnect":
        raise PubSubError("broadcaster disconnected")
    if "bytes" not in raw_result:
        raise PubSubError("unexpected message type received")

    payload = cast(WSMessageBytes, raw_result)["bytes"]
    state.read_task = make_websocket_read_task(state.websocket)

    reader = io.BytesIO(payload)
    prefix = parse_b2s_message_prefix(reader)

    message = B2S_AnyMessageParser.parse(prefix.flags, prefix.type, reader)
    check_read_message(state, message)
    return CheckResult.RESTART



import asyncio
import random
import sys
import time
from typing import Any, List, Set

import aiohttp

from lonelypsc.config.ws_config import WebsocketPubSubConfig
from lonelypsc.util.errors import combine_multiple_exceptions
from lonelypsc.ws.internal_callbacks import finalize_internal_callback
from lonelypsc.ws.state import (
    InternalMessageStateDroppedSent,
    InternalMessageStateDroppedUnsent,
    InternalMessageStateType,
    RetryInformation,
    State,
    StateClosed,
    StateConnecting,
    StateType,
    StateWaitingRetry,
    TasksOnceOpen,
)
from lonelypsc.ws.websocket_connect_task import make_websocket_connect_task

if sys.version_info >= (3, 11):
    from typing import Never
else:
    from typing import NoReturn as Never


async def handle_connection_failure(
    *,
    config: WebsocketPubSubConfig,
    cancel_requested: asyncio.Event,
    retry: RetryInformation,
    tasks: TasksOnceOpen,
    exception: BaseException,
    backgrounded: Set[asyncio.Task[Any]]
) -> State:
    """Handles a connection failure by either moving to the next broadcaster,
    moving to WAITING_RETRY, or moving to CLOSED.

    This assumes:

    - all necessary cleanup for the previous connection is already completed
    - suppressing the exception if retrying is not an issue

    Args:
        config (WebsocketPubSubConfig): The configuration for the subscriber
        cancel_requested (asyncio.Event): whether the state machine is trying to
            gracefully shutdown (set) or not (not set)
        retry (RetryInformation): how to determine the next broadcaster
        tasks (TasksOnceOpen): the tasks that need to be performed if a broadcaster
            can be reached or canceled if moving to closed
        exception (BaseException): the exception that caused the connection failure;
            will be included somewhere in the error if no retries are possible
        backgrounded (Set[asyncio.Task[Any]]): the set of backgrounded tasks that
            if they fail it must be treated as irrecoverable, but whose result otherwise
            is unimportant. these are assumed to be like notification callbacks in the
            sense that they don't need cancellation and should be called even after an
            interrupt before shutting down

    Returns:
        the new state for the state machine

    Raises:
        BaseException: if no more retries can be attempted, raises the exception
            instead of directly returning StateClosed; this should eventually be
            caught and transition to StateClosed before raising the connection all
            the way to outside this library
    """

    if cancel_requested.is_set():
        await cleanup_tasks_and_raise_on_error(tasks, backgrounded, "cancel requested")
        return StateClosed(type=StateType.CLOSED)

    new_backgrounded = set()
    for bknd in backgrounded:
        if not bknd.done():
            new_backgrounded.add(bknd)
            continue

        if bknd.exception() is not None:
            await cleanup_tasks_and_raise(
                tasks,
                backgrounded,
                "backgrounded sweep failed",
                Exception("backgrounded task failed"),
            )

    try:
        next_broadcaster = next(retry.iterator)
        client_session = aiohttp.ClientSession()
        return StateConnecting(
            type=StateType.CONNECTING,
            config=config,
            client_session=client_session,
            websocket_task=make_websocket_connect_task(
                config, next_broadcaster, client_session
            ),
            cancel_requested=cancel_requested,
            broadcaster=next_broadcaster,
            retry=retry,
            tasks=tasks,
            backgrounded=new_backgrounded,
        )
    except StopIteration:
        ...

    if retry.iteration < config.outgoing_initial_connect_retries:
        retry.iteration += 1
        retry.iterator = iter(retry.shuffler)

        return StateWaitingRetry(
            type=StateType.WAITING_RETRY,
            config=config,
            cancel_requested=cancel_requested,
            retry=retry,
            tasks=tasks,
            retry_at=time.time() + (2 ** (retry.iteration - 1) + random.random()),
            backgrounded=new_backgrounded,
        )

    await cleanup_tasks_and_raise(
        tasks, new_backgrounded, "retries exhausted", exception
    )


async def cleanup_tasks_and_return_errors(
    tasks: TasksOnceOpen, backgrounded: Set[asyncio.Task[Any]]
) -> List[BaseException]:
    """Cleans up the given tasks, returning any errors that occurred"""
    cleanup_excs: List[BaseException] = []
    while tasks.resending_notifications:
        notif = tasks.resending_notifications.pop()
        backgrounded.add(
            asyncio.create_task(
                finalize_internal_callback(
                    notif.callback,
                    InternalMessageStateDroppedSent(
                        type=InternalMessageStateType.DROPPED_SENT
                    ),
                )
            )
        )

    for notif in tasks.unsent_notifications.drain():
        backgrounded.add(
            asyncio.create_task(
                finalize_internal_callback(
                    notif.callback,
                    InternalMessageStateDroppedUnsent(
                        type=InternalMessageStateType.DROPPED_UNSENT
                    ),
                )
            )
        )

    for bknd in backgrounded:
        try:
            await bknd
        except BaseException as e:
            cleanup_excs.append(e)

    return cleanup_excs


async def cleanup_tasks_and_raise_on_error(
    tasks: TasksOnceOpen, backgrounded: Set[asyncio.Task[Any]], message: str
) -> None:
    """Cleans up the given tasks and raises an exception if any errors occurred"""
    cleanup_excs = await cleanup_tasks_and_return_errors(tasks, backgrounded)

    if cleanup_excs:
        raise combine_multiple_exceptions(message, cleanup_excs)


async def cleanup_tasks_and_raise(
    tasks: TasksOnceOpen,
    backgrounded: Set[asyncio.Task[Any]],
    message: str,
    cause: BaseException,
) -> Never:
    """Cleans up the given tasks list and raises the given exception;
    if closing the tasks raises an exception, that exception is combined
    with the original exception and raised
    """
    cleanup_excs = await cleanup_tasks_and_return_errors(tasks, backgrounded)

    if cleanup_excs:
        raise combine_multiple_exceptions(
            message,
            cleanup_excs,
            context=cause,
        )

    raise cause



import asyncio
from contextlib import asynccontextmanager
from typing import AsyncGenerator, List, Union

from lonelypsc.util.errors import combine_multiple_exceptions
from lonelypsc.ws.state import (
    InternalMessage,
    InternalMessageState,
    InternalMessageStateAcknowledged,
    InternalMessageStateAndCallback,
    InternalMessageStateDroppedSent,
    InternalMessageStateDroppedUnsent,
    InternalMessageStateResending,
    InternalMessageStateSent,
    InternalMessageStateType,
    InternalMessageStateUnsent,
)


def _is_intermediate_state(state: InternalMessageState) -> bool:
    return state.type in (
        InternalMessageStateType.UNSENT,
        InternalMessageStateType.SENT,
        InternalMessageStateType.RESENDING,
    )


def sweep_internal_message(message: InternalMessage) -> None:
    """The active management referenced in `inform_internal_message`; checks on the
    progress of the state callback, raising an exception if it completed with an
    error and otherwise moving it closer to the best known state for the message.
    """
    if message.callback.task is not None and message.callback.task.done():
        finished = message.callback.task
        message.callback.task = None
        finished.result()  # raises exception if there was one

    if message.callback.queued is not None and message.callback.task is None:
        assert _is_intermediate_state(message.callback.state)
        message.callback.state = message.callback.queued
        message.callback.queued = None
        message.callback.task = asyncio.create_task(
            message.callback.callback(message.callback.state)
        )


def inform_internal_message(
    message: InternalMessage,
    state: Union[
        InternalMessageStateUnsent,
        InternalMessageStateSent,
        InternalMessageStateResending,
    ],
) -> None:
    """Queues a task to be scheduled to inform the callback of the new state of the
    message.

    In order for this to work, the internal message must still be actively
    monitored until it reaches a final state by continuously calling
    sweep_internal_message until the final state is reached, at which point
    finalize_internal_callback can be scheduled on the event loop and tracked to
    completion

    Args:
        message (InternalMessage): the message to inform
        state (InternalMessageState): the non-final state to inform the message of
    """

    if message.callback.state.type == state.type:
        message.callback.queued = None
        return

    message.callback.queued = state


@asynccontextmanager
async def readable_internal_message(
    message: InternalMessage,
) -> AsyncGenerator[None, None]:
    """Acts as an asynchronous context manager that, while it is active,
    the messages stream is readable (either in UNSENT or RESENDING)

    Raises an exception if this is not possible through valid transitions
    (i.e., the message is not in UNSENT, SENT, or RESENDING)
    """

    message.callback.queued = None

    while True:
        assert _is_intermediate_state(message.callback.state), "final state reached"
        if message.callback.task is None:
            break

        await message.callback.task
        message.callback.queued = None
        sweep_internal_message(message)

    future: asyncio.Future[None] = asyncio.get_running_loop().create_future()

    async def _task_target() -> None:
        await future

    try:
        message.callback.task = asyncio.create_task(_task_target())

        if message.callback.state.type == InternalMessageStateType.SENT:
            message.callback.state = InternalMessageStateResending(
                type=InternalMessageStateType.RESENDING,
            )
            await message.callback.callback(message.callback.state)

        assert message.callback.state.type in (
            InternalMessageStateType.UNSENT,
            InternalMessageStateType.RESENDING,
        )
        message.callback.queued = None
        yield
    finally:
        future.set_result(None)


async def finalize_internal_callback(
    callback: InternalMessageStateAndCallback,
    final_state: Union[
        InternalMessageStateAcknowledged,
        InternalMessageStateDroppedUnsent,
        InternalMessageStateDroppedSent,
    ],
) -> None:
    """Returns a coroutine that runs until the callback has been notified of
    its final state; consumes errors on the way, raising them at the end.

    Args:
        callback (InternalMessageStateAndCallback): the callback to finalize
        final_state (InternalMessageState (final)): the final state of the message
    """

    exceptions: List[BaseException] = []
    while callback.task is not None:
        if callback.task.done():
            if (exc := callback.task.exception()) is not None:
                exceptions.append(exc)
            callback.task = None
            break

        try:
            await callback.task
            # NOTE: callback.task may have changed here as we may not have
            # been the first thing scheduled
        except BaseException:
            ...  # will see it in the next iteration if still relevant

    assert _is_intermediate_state(callback.state), "final state already reached"
    callback.state = final_state  # prevents future tasks from being scheduled

    try:
        await callback.callback(final_state)
    except BaseException as e:
        exceptions.append(e)

    if exceptions:
        raise combine_multiple_exceptions("error while finalizing callback", exceptions)



import asyncio
import warnings
from typing import (
    Any,
    Generic,
    Iterable,
    List,
    Set,
    TypeVar,
    cast,
)

import uvicorn
from fastapi import APIRouter, FastAPI
from lonelypsp.util.cancel_and_check import cancel_and_check

from lonelypsc.config.http_config import (
    HttpPubSubBindManualConfig,
    HttpPubSubBindUvicornConfig,
)

T = TypeVar("T")


class _EmptyEventSet(Generic[T]):
    """Looks like a set(), but when it empties it sets the event"""

    def __init__(self, raw: Set[T]) -> None:
        self.raw = raw
        self.event = asyncio.Event()

        if not raw:
            self.event.set()

    # in order from https://docs.python.org/3/library/stdtypes.html#set
    def __len__(self) -> int:
        return len(self.raw)

    def __contains__(self, item: T) -> bool:
        return item in self.raw

    def isdisjoint(self, other: Iterable[T]) -> bool:
        return self.raw.isdisjoint(other)

    def issubset(self, other: Iterable[T]) -> bool:
        return self.raw.issubset(other)

    # order
    def __eq__(self, other: Any) -> bool:
        return self.raw == other

    def __ne__(self, other: Any) -> bool:
        return self.raw != other

    def __lt__(self, other: Set[T]) -> bool:
        return self.raw < other

    def __le__(self, other: Set[T]) -> bool:
        return self.raw <= other

    def __gt__(self, other: Set[T]) -> bool:
        return self.raw > other

    def __ge__(self, other: Set[T]) -> bool:
        return self.raw >= other

    def union(self, *others: Iterable[T]) -> Set[T]:
        return self.raw.union(*others)

    def __or__(self, other: Set[T]) -> Set[T]:
        return self.raw | other

    def intersection(self, *others: Iterable[T]) -> Set[T]:
        return self.raw.intersection(*others)

    def __and__(self, other: Set[T]) -> Set[T]:
        return self.raw & other

    def difference(self, *others: Iterable[T]) -> Set[T]:
        return self.raw.difference(*others)

    def __sub__(self, other: Set[T]) -> Set[T]:
        return self.raw - other

    def symmetric_difference(self, other: Iterable[T]) -> Set[T]:
        return self.raw.symmetric_difference(other)

    def __xor__(self, other: Set[T]) -> Set[T]:
        return self.raw ^ other

    def copy(self) -> Set[T]:
        return self.raw.copy()

    def update(self, *others: Iterable[T]) -> None:
        self.raw.update(*others)
        if not self.raw:
            self.event.set()

    def __ior__(self, other: Set[T]) -> None:
        self.raw |= other
        if not self.raw:
            self.event.set()

    def intersection_update(self, *others: Iterable[T]) -> None:
        self.raw.intersection_update(*others)
        if not self.raw:
            self.event.set()

    def __iand__(self, other: Set[T]) -> None:
        self.raw &= other
        if not self.raw:
            self.event.set()

    def difference_update(self, *others: Iterable[T]) -> None:
        self.raw.difference_update(*others)
        if not self.raw:
            self.event.set()

    def __isub__(self, other: Set[T]) -> None:
        self.raw -= other
        if not self.raw:
            self.event.set()

    def symmetric_difference_update(self, other: Iterable[T]) -> None:
        self.raw.symmetric_difference_update(other)
        if not self.raw:
            self.event.set()

    def __ixor__(self, other: Set[T]) -> None:
        self.raw ^= other
        if not self.raw:
            self.event.set()

    def add(self, elem: T) -> None:
        self.raw.add(elem)

    def remove(self, elem: T) -> None:
        self.raw.remove(elem)
        if not self.raw:
            self.event.set()

    def discard(self, elem: T) -> None:
        self.raw.discard(elem)
        if not self.raw:
            self.event.set()

    def pop(self) -> T:
        elem = self.raw.pop()
        if not self.raw:
            self.event.set()
        return elem

    def clear(self) -> None:
        self.raw.clear()
        self.event.set()


class BindWithUvicornCallback:
    """Fulfills the HttpPubSubBindManualCallback using uvicorn as the runner"""

    def __init__(self, settings: HttpPubSubBindUvicornConfig):
        self.settings = settings

    async def _shutdown(self, uv_server: uvicorn.Server) -> None:
        """Gracefully shutdown the server without polling; must faster than
        uv_server.shutdown()
        """
        for server in uv_server.servers:
            server.close()

        for connection in list(uv_server.server_state.connections):
            connection.shutdown()

        timeout_task = (
            asyncio.create_task(
                asyncio.sleep(uv_server.config.timeout_graceful_shutdown)
            )
            if uv_server.config.timeout_graceful_shutdown is not None
            else None
        )

        connection_done_tasks: List[asyncio.Task[Any]] = []

        if uv_server.server_state.connections:
            replacer = _EmptyEventSet(uv_server.server_state.connections)
            for conn in replacer.raw:
                conn.connections = cast(Any, replacer)
            uv_server.server_state.connections = cast(Any, replacer)
            connection_done_tasks.append(asyncio.create_task(replacer.event.wait()))

        if connection_done_tasks or uv_server.server_state.tasks:
            graceful_shutdown_complete_task = asyncio.create_task(
                asyncio.wait(
                    [
                        *connection_done_tasks,
                        *uv_server.server_state.tasks,
                    ],
                    return_when=asyncio.ALL_COMPLETED,
                )
            )
            if timeout_task is not None:
                await asyncio.wait(
                    [
                        timeout_task,
                        graceful_shutdown_complete_task,
                    ],
                    return_when=asyncio.FIRST_COMPLETED,
                )
                if await cancel_and_check(timeout_task, True):
                    warnings.warn_explicit(
                        "graceful shutdown timed out", ResourceWarning, "", 0
                    )
                await cancel_and_check(graceful_shutdown_complete_task)
            else:
                await graceful_shutdown_complete_task
                assert not uv_server.server_state.connections
                assert not uv_server.server_state.tasks

            for task in uv_server.server_state.tasks:
                await cancel_and_check(task)

            for task in connection_done_tasks:
                await cancel_and_check(task)

        if not uv_server.force_exit:
            await uv_server.lifespan.shutdown()

    async def _serve(self, server: uvicorn.Server, cancel_event: asyncio.Event) -> None:
        # if canceled before starting, nothing to do
        if cancel_event.is_set():
            return

        # setup lifespan
        config = server.config
        if not config.loaded:
            config.load()

        server.lifespan = config.lifespan_class(config)

        # don't interrupt startup; too likely to leak
        await server.startup()

        # lifespan events aborted
        if server.should_exit:
            return

        # if canceled, go straight to shutdown without starting main loop
        if cancel_event.is_set():
            # don't interrupt shutdown; too likely to leak
            await self._shutdown(server)
            return

        # server.main_loop relies on sleeping but isn't safe to cancel because
        # it might be in on_tick if unlucky; so we reimplement here
        cancel_event_wait_task = asyncio.create_task(cancel_event.wait())

        counter = 0
        while not await server.on_tick(counter):
            counter += 1
            if counter == 864000:
                counter = 0

            sleep_task = asyncio.create_task(asyncio.sleep(0.1))
            await asyncio.wait(
                [sleep_task, cancel_event_wait_task],
                return_when=asyncio.FIRST_COMPLETED,
            )
            if await cancel_and_check(sleep_task, True):
                # since sleep_task wasn't done, must have been canceled
                assert cancel_event_wait_task.done()
                break

        await self._shutdown(server)

    async def __call__(self, router: APIRouter) -> None:
        app = FastAPI()
        app.include_router(router)
        app.router.redirect_slashes = False
        uv_config = uvicorn.Config(
            app,
            host=self.settings["host"],
            port=self.settings["port"],
            lifespan="off",
            # prevents spurious cancellation errors
            log_level="warning",
            # reduce default logging since this isn't the main deal for the process
        )
        uv_server = uvicorn.Server(uv_config)
        cancel_event = asyncio.Event()
        serve_task = asyncio.create_task(self._serve(uv_server, cancel_event))

        try:
            await asyncio.shield(serve_task)
        finally:
            cancel_event.set()
            await serve_task


async def handle_bind_with_uvicorn(
    settings: HttpPubSubBindUvicornConfig,
) -> HttpPubSubBindManualConfig:
    """Converts the bind with uvicorn settings into the generic manual config"""
    return {
        "type": "manual",
        "callback": BindWithUvicornCallback(settings),
    }



import time

from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.state import (
    OpenRetryInformationStable,
    OpenRetryInformationType,
    StateOpen,
)


def check_retry_tentative(state: StateOpen) -> CheckResult:
    """Tries to change the states stability level from tentative to stable,
    returning CONTINUE if it cannot and RESTART if it does
    """
    if state.retry.type != OpenRetryInformationType.TENTATIVE:
        return CheckResult.CONTINUE

    if state.retry.stable_at > time.time():
        return CheckResult.CONTINUE

    state.retry = OpenRetryInformationStable(type=OpenRetryInformationType.STABLE)
    return CheckResult.RESTART



import time

from lonelypsp.stateful.constants import SubscriberToBroadcasterStatefulMessageType
from lonelypsp.stateful.messages.continue_receive import (
    S2B_ContinueReceive,
    serialize_s2b_continue_receive,
)

from lonelypsc.ws.handlers.open.websocket_url import (
    make_for_send_websocket_url_and_change_counter,
)
from lonelypsc.ws.state import StateOpen, UnsentAckContinueReceive


async def send_continue_receive(
    state: StateOpen, message: UnsentAckContinueReceive
) -> None:
    """Produces the required authorization header and sends the continue receive message"""

    url = make_for_send_websocket_url_and_change_counter(state)
    tracing = b""  # TODO: tracing
    authorization = await state.config.authorize_stateful_continue_receive(
        tracing=tracing,
        identifier=message.identifier,
        part_id=message.part_id,
        url=url,
        now=time.time(),
    )
    serd_message = serialize_s2b_continue_receive(
        S2B_ContinueReceive(
            type=SubscriberToBroadcasterStatefulMessageType.CONTINUE_RECEIVE,
            identifier=message.identifier,
            part_id=message.part_id,
            authorization=authorization,
            tracing=tracing,
        ),
        minimal_headers=state.config.websocket_minimal_headers,
    )
    await state.websocket.send_bytes(serd_message)



from dataclasses import dataclass
from typing import TYPE_CHECKING, Type

from lonelypsc.client import PubSubNotifyResult


@dataclass
class HttpPubSubNotifyResult:
    notified: int


if TYPE_CHECKING:
    _: Type[PubSubNotifyResult] = HttpPubSubNotifyResult
